import tla._ast
import tla._lre as _lr


source = r'''
---- MODULE example ----
A == """
    This is a multiline string
    in TLA+
    """
========================
'''


def test_multiline_strings():
    parser = _lr.Parser()
    tree = parser.parse(source)
    print(tree)
    out = tla._ast.pprint_tree(tree)
    print(out)


if __name__ == '__main__':
    test_multiline_strings()



"""Syntax tree that represents TLA+."""
import argparse as _arg
import collections as _cl
import collections.abc as _abc
import enum
import functools as _ft
import itertools as _itr
import pprint as _pp
import shlex as _sh
import subprocess as _sbp
import time
import typing as _ty

import tla._lex as _lex
import tla._langdef as _opdef


_MIN_LINE_LEN = 4
_DASH_LINE = 'DASH_LINE'
_EQ_LINE = 'EQ_LINE'
_LPAREN = 'LPAREN'
_RPAREN = 'RPAREN'
_LBRACE = 'LBRACE'
_RBRACE = 'RBRACE'
_LBRACKET = 'LBRACKET'
_RBRACKET = 'RBRACKET'
_TLA_OPERATOR_FIXITY = _opdef.TLA_OPERATOR_FIXITY
_lexemes_to_tokens = _lex.map_lexemes_to_tokens()


NODE_SPECS = dict(
    AXIOM=
        'name, expression, axiom_keyword',
    BOOLEAN_LITERAL=
        'value',
    BY=
        'only, facts, names',
    CASES=
        'cases, other',
        # Represents both `CASE` and `ELIF`
        # constructs. The specific form is
        # considered concrete information
        # (tokeny).
    CASE_ITEM=
        'predicate, expression',
    CHOOSE=
        'declaration, predicate',
        # "predicate" here is for readability:
        # this expression may take values outside
        # of the set `BOOLEAN`
    CONSTANTS=
        'names',
    DEFINE=
        'definitions',
    EXCEPT=
        'function, changes',
    FAIRNESS=
        'operator, subscript, action',
    FIELD=
        'expression, name',
    FLOAT_NUMERAL=
        'value',
    FUNCTION=
        'declaration, value',
    FUNCTION_APPLICATION=
        'function, arguments',
    FUNCTION_CHANGE=
        'item, expression',
    HAVE=
        'expression',
    HIDE=
        'facts, names',
    IF=
        'predicate, then, else_',
    IFS=
        'cases, other',
    IF_ITEM=
        'predicate, expression',
    INSTANCE=
        'name, with_substitution, local',
    INTEGRAL_NUMERAL=
        'value',
    KEY_BOUND=
        'name, bound',
    LAMBDA=
        'parameters, expression',
    LET=
        'definitions, expression',
    MODULE=
        'name, extendees, units',
    MODULE_NAME=
        'name',
    OBVIOUS=
        '',
    OMITTED=
        '',
    OPERATOR_APPLICATION=
        'operator, arguments',
    OPERATOR_DECLARATION=
        'name, level, new_keyword',
    OPERATOR_DEFINITION=
        'name, arity, '
        'definiens, local, function',
            # `function` means
            # the form:
            # `f[...] == ...`
    PARAMETER_DECLARATION=
        'name, arity',
    PARENTHESES=
        'expression',
    PICK=
        'declarations, predicate',
    PROOF=
        'steps, proof_keyword',
    PROOF_CASE=
        'expression',
    PROOF_STEP=
        'name, main, proof, proof_keyword',
        # "proof step" distinguishes
        # from a pair of states in
        # the semantics
    QUANTIFICATION=
        'quantifier, declarations, predicate',
        # "predicate" here is for readability:
        # this expression may take values outside
        # of the set `BOOLEAN`
    QUANTIFIER_WITNESS=
        'expressions',
    RECORD=
        'key_values',
    RECURSIVE_OPERATORS=
        'names',
    SEQUENT=
        'assume, prove',
    SET_COMPREHENSION=
        'item, declarations',
        # expression ?
    SET_ENUMERATION=
        'items',
        # expressions ?
    SET_OF_BOOLEANS=
        '',
    SET_OF_FUNCTIONS=
        'domain, codomain',
    SET_OF_RECORDS=
        'key_bounds',
    SET_OF_STRINGS=
        '',
    SET_SLICE=
        'declaration, predicate',
    STRING_LITERAL=
        'value',
    SUBEXPRESSION_REFERENCE=
        'items',
    SUBSCRIPTED_ACTION=
        'operator, action, subscript',
    SUFFICES=
        'predicate',
    TAKE=
        'declarations',
    TEMPORAL_QUANTIFICATION=
        'quantifier, declarations, predicate',
        # "predicate" here is for readability:
        # this expression may take values outside
        # of the set `BOOLEAN`
    THEOREM=
        'name, goal, proof, theorem_keyword',
    TUPLE=
        'items',
    USE=
        'facts, names, only',
    VARIABLES=
        'names',
    VERTICAL_LIST=
        'operator, arguments',
    VERTICAL_ITEM=
        'operator, expression',
    WITH_INSTANTIATION=
        'name, expression'
    )


NodeTypes = enum.StrEnum(
    'NodeTypes',
    sorted(NODE_SPECS))


def make_nodes(
        ) -> _arg.Namespace:
    """Return syntax tree classes."""
    node_types = dict()
    items = NODE_SPECS.items()
    for node_name, attr_names in items:
        class_name = enum_to_class(node_name)
        node_enum = NodeTypes[node_name]
        # attribute names
        attr_names = [
            attr
            for attr in
                attr_names.split(',')
            if attr]
        attr_names = (
            'symbol',
            *attr_names,
            'node_index')
        # default values of attributes
        n_attrs = len(attr_names)
        defaults = (
            (node_enum,) +
            (None,) * (n_attrs - 1))
        attrs = ', '.join(attr_names)
        node_types[class_name] = _cl.namedtuple(
            class_name, attrs,
            defaults=defaults)
    return _arg.Namespace(
        **node_types)


def enum_to_class(
        enum_name:
            str
        ) -> str:
    """Return class name."""
    return enum_name.title().replace('_', '')


def class_to_enum(
        class_name:
            str
        ) -> str:
    """Return enumeration name."""
    def prepend(
            char:
                str
            ) -> str:
        if char.isupper():
            return f'_{char}'
        return char
    enum_name = ''.join(map(
        prepend, class_name))
    return enum_name[1:].upper()


_nodes = make_nodes()
_LEAF_PROOFS = {
    'BY',
    'OBVIOUS',
    'OMITTED'}


def map_steps_to_proofs(
        tree):
    """Map to nested proofs.

    Based on step numbers, lists of steps
    are mapped to hierarchical trees of
    proofs.
    """
    rec = map_steps_to_proofs
    match tree:
        case str() | bool() | None:
            return tree
        case tuple() if not hasattr(tree, 'symbol'):
            return tuple(map(rec, tree))
        case list():
            return list(map(rec, tree))
    enum_name = enum_to_class(
        tree.symbol.name)
    match enum_name:
        case 'Theorem':
            has_leaf_proof = (
                tree.proof is not None and
                len(tree.proof.steps) == 1 and
                tree.proof.steps[0].symbol.name in
                    _LEAF_PROOFS)
            if has_leaf_proof:
                proof = tree.proof
            else:
                proof = rec(tree.proof)
            return _nodes.Theorem(
                name=tree.name,
                goal=tree.goal,
                proof=proof,
                theorem_keyword=
                    tree.theorem_keyword)
        case 'ProofStep':
            proof = rec(tree.proof)
            return _nodes.ProofStep(
                name=tree.name,
                main=tree.main,
                proof=proof,
                proof_keyword=
                    tree.proof_keyword)
        case 'Proof':
            return make_steps_tree(tree)
        case _:
            symbol, *attrs = tree
            attrs = map(rec, attrs)
            cls = type(tree)
            return cls(symbol, *attrs)


def make_steps_tree(
        tree):
    """Return proof tree, from steps sequence."""
    steps_stack = [list()]
    current_level = 1
    steps = tree.steps
    end_step = _nodes.ProofStep(
        name='<1>')
        # used to cause the final popping
        # of levels in `steps_stack` into
        # the first level
    steps.append(end_step)
    for step in steps:
        proof = step.proof
        has_proof = (
            proof is not None and
            proof.symbol.name not in
                _LEAF_PROOFS)
        if has_proof:
            raise ValueError(
                'proof step already has '
                'nonleaf proof')
        level = _step_level(step)
        # print(f'{level = }')
        if level > current_level:
            # print('adding proof level')
            steps_stack.append(list())
            current_level = level
        elif level < current_level:
            while level < current_level:
                # print('popping proof level')
                proof_steps = steps_stack.pop()
                # get step of smaller level
                prev_level = steps_stack[-1]
                last_step = prev_level.pop()
                if last_step.proof is not None:
                    step_num = last_step.name
                    raise AssertionError(
                        f'`{step_num}` already has '
                        'a proof, that would be '
                        'overwritten by this '
                        'sequence of proof-steps.')
                current_level = _step_level(last_step)
                proof = _nodes.Proof(
                    steps=proof_steps)
                last_step = _nodes.ProofStep(
                    name=last_step.name,
                    main=last_step.main,
                    proof=proof,
                    proof_keyword=
                        last_step.proof_keyword)
                steps_stack[-1].append(
                    last_step)
        if level != current_level:
            raise AssertionError(
                level, current_level)
        steps_stack[-1].append(step)
    if len(steps_stack) > 1:
        raise AssertionError(
            len(steps_stack))
    proof_steps, = steps_stack
    proof_steps.pop()  # end-step
    return _nodes.Proof(
        steps=proof_steps,
        proof_keyword=tree.proof_keyword)


def _step_level(
        step
        ) -> int:
    """Return level of `step`."""
    label = step.name
    prefix, _ = label.split('>')
    level = int(prefix.lstrip('<'))
    return level


def op_decl_name_bound_arity(
        tree
        ) -> tuple:
    """Return details of `OperatorDeclaration`."""
    if tree.operator == '\\in':
        name, bound = tree.arguments
        arity = None
    else:
        name = tree.operator
        bound = None
        arity = tree.arguments
    if name.arguments is not None:
        raise AssertionError(name)
    name = name.operator
    return (
        name,
        bound,
        arity)


def declarations_as_name_bounds(
        expressions:
            list[tuple]
        ) -> list[tuple]:
    r"""Return names and bounds of declared variables.

    Also duplicates bounds, for example
    `x, y \in S` becomes
    `[(x, S), `(y, S)]`.
    """
    name_bounds = list()
    for expr in expressions:
        if expr.operator == '\\in':
            name, bound = expr.arguments
            name_bounds.append(
                (name, bound))
                # `name` can be a `Tuple`
        elif expr.arguments is None:
            name = expr
            bound = None
            name_bounds.append(
                (name, bound))
        else:
            raise ValueError(expr)
    # ditto bounds
    ditto_bounds = list()
    last_bound = None
    for name, bound in reversed(name_bounds):
        if bound is None:
            bound = last_bound
        last_bound = bound
        ditto_bounds.append(
            (name, bound))
    return list(reversed(ditto_bounds))


def _pformat_nt_tree(
        tree:
            tuple |
            list |
            str
        ) -> str:
    """Format `tree`."""
    d = _pformat_nt_tree_recurse(tree)
    return _pp.pformat(d, width=70)


def _pformat_nt_tree_recurse(
        tree):
    if (isinstance(tree, tuple) and
            hasattr(tree, '_fields')):
        return _pformat_namedtuple(tree)
    elif isinstance(tree, (tuple, list)):
        return _pformat_list(tree)
    else:
        return tree


def _pformat_namedtuple(
        nt):
    d = dict()
    for key in nt._fields:
        value = getattr(nt, key)
        d[key] = _pformat_nt_tree_recurse(value)
    return d


def _pformat_list(
        x):
    return [
        _pformat_nt_tree_recurse(v)
        for v in x]


def dump_tree_as_graph(
        tree:
            tuple,
        filename:
            str
        ) -> None:
    """Layout syntax tree as PDF."""
    graph = _tree_to_graph(tree)
    dot_text = list()
    for node, label in graph.nodes.items():
        node_dot = f'{node} [label="{label}"];'
        dot_text.append(node_dot)
    for node, other in graph.edges:
        edge_dot = f'{node} -> {other};'
        dot_text.append(edge_dot)
    dot_text = '\n'.join(dot_text)
    dot_text = f'digraph {{ {dot_text} }}'
    # write PDF file
    pdf_filename = f'{filename}.pdf'
    dot_filename = f'{filename}.dot'
    with open(dot_filename, 'w') as fd:
        fd.write(dot_text)
    prog = _sh.split(f'''
        dot -Tpdf
            -o {pdf_filename}
            {dot_filename}
        ''')
    _sbp.call(prog)


class DiGraph:
    """Graph with directed edges."""

    def __init__(
            self
            ) -> None:
        self.nodes = dict()
        self.edges = set()


def _tree_to_graph(
        tree):
    """Convert tree to graph."""
    graph = DiGraph()
    _tree_to_graph_recurse(tree, graph)
    return graph


def _tree_to_graph_recurse(
        tree:
            tuple |
            list |
            str,
        graph:
            DiGraph
        ) -> int:
    """Create graph node for `tree`."""
    u = len(graph.nodes)
    if isinstance(tree, tuple):
        label = f'{tree.symbol.name}'
        label = label.replace(
            '.', r'_').replace(
            '"', '').replace(
            '\\', '\\\\')
        graph.nodes[u] = label
    elif isinstance(tree, str):
        label = tree
        label = label.replace(
            '.', r'_').replace(
            '"', '').replace(
            '\\', '\\\\')
        graph.nodes[u] = label
    if isinstance(tree, (tuple, list)):
        for succ in tree:
            if succ is None:
                continue
            v = _tree_to_graph_recurse(
                succ, graph)
            graph.edges.add((u, v))
    return u


Token = _cl.namedtuple(
    'Token',
    'symbol, value, '
    'start_line, start_column, '
    'end_line, end_column, '
    'filename, ast_node_index',
        # symbol \in NodeTypes
        # value \in STRING
        # start = location of start of lexeme
        #     as number of characters from
        #     start of file
        # start \in STRING
        # end = location of end of lexeme
        #     as number of characters from
        #     start of file
        # end \in STRING
        # filename \in STRING
    defaults=(
        None, None, None, None,
        None, None, None, None)
        # defaults enable having undefined
        # location for tokens created when
        # concretizing a syntax tree,
        # before positional information
        # has yet been copied from
        # the input lexemes.
    )


def _make_token(
        value):
    return Token(value, value)


def _make_name_token(
        name):
    if not isinstance(name, str):
        raise AssertionError(name)
    return Token('NAME', name)


def tokenize_ast_root(
        tree:
            tuple,
        _node_index:
            int=0
        ) -> _abc.Iterator[
            Token]:
    """Yield tokens of syntax `tree`.

    Nodes are indexed by DFS preorder.

    The intention is to copy information
    from tokens of the input sequence,
    e.g., positional information,
    without changing the parser.
    """
    tree = _index_by_dfs(tree)
    yield from _tokenize_ast(tree)


def _tokenize_ast(
        tree:
            tuple
        ) -> _abc.Iterator[
            Token]:
    for token in _tokenize_ast_recurse(tree):
        if token.ast_node_index is not None:
            yield token
            continue
        if hasattr(tree, 'node_index'):
            node_index = tree.node_index
        else:
            node_index = None
        token = Token(
            symbol=token.symbol,
            value=token.value,
            start_line=token.start_line,
            start_column=token.start_column,
            end_line=token.end_line,
            end_column=token.end_column,
            filename=token.filename,
            ast_node_index=node_index)
        yield token


def _tokenize_ast_recurse(
        tree:
            tuple
        ) -> _abc.Iterator[
            Token]:
    """Yield tokens for `tree`."""
    match tree:
        case '_':
            yield Token('UNDERSCORE', '_')
            return
        case str() | bool() | None:
            raise AssertionError(tree)
        case list():
            raise AssertionError(tree)
    match enum_to_class(tree.symbol.name):
        case 'Axiom':
            yield _make_token('AXIOM')
            if tree.name:
                yield _make_name_token(tree.name)
                yield Token('DOUBLE_EQ', '==')
            yield from _tokenize_ast(
                tree.expression)
        case 'BooleanLiteral':
            yield _make_token(tree.value)
        case 'By':
            yield _make_token('BY')
            if tree.only:
                yield _make_token('ONLY')
            yield from _tokenize_comma_list(
                tree.facts)
            if tree.names:
                yield _make_token('DEF')
            yield from _tokenize_names(tree.names)
        case 'Cases':
            # for converting to an `ELIF` form
            # at the token level, map:
            # - `CASE` to `IF`,
            # - `->` to `THEN`,
            # - the pair `[]` `OTHER` to `ELSE`
            # - `[]` to `ELIF`
            yield _make_token('CASE')
            gen = _itr.chain.from_iterable(
                _tokenize_ast(case)
                for case in tree.cases)
            yield from gen
            yield _make_token('OTHER')
            yield from _tokenize_ast(tree.other)
        case 'CaseItem':
            yield from _tokenize_ast(tree.predicate)
            yield _make_token('RARROW', '->')
            yield from _tokenize_ast(tree.expression)
            yield _make_token('LBRACKET_RBRACKET', '[]')
        case 'Choose':
            yield _make_token('CHOOSE')
            _tokenize_ast(tree.declaration)
            yield Token('COLON', ':')
            yield from _tokenize_ast(
                tree.predicate)
        case 'Constants':
            yield _make_token('CONSTANT')
            yield from _tokenize_names(tree.names)
        case 'Define':
            yield _make_token('DEFINE')
            _tokenize_list(tree.definitions)
        case 'Except':
            yield Token(_LBRACKET, '[')
            yield from _tokenize_ast(tree.function)
            yield _make_token('EXCEPT')
            yield from _tokenize_comma_list(
                tree.changes)
            yield Token(_RBRACKET, ']')
        case 'Fairness':
            yield Token('WF', 'WF_')
            yield from _tokenize_ast(teee.subscript)
            yield Token(_LPAREN, '(')
            yield from _tokenize_ast(tree.action)
            yield Token(_RPAREN, ')')
        case 'Field':
            yield from _tokenize_ast(tree.expression)
            yield Token('DOT', '.')
            yield Token('NAME', tree.name)
        case 'FloatNumeral':
            yield Token('FLOAT', tree.value)
        case 'Function':
            yield Token(_LBRACKET, '[')
            yield from _tokenize_comma_list(
                tree.declaration)
            yield Token('MAPSTO', '|->')
            yield from _tokenize_ast(tree.value)
            yield Token(_RBRACKET, ']')
        case 'FunctionApplication':
            yield from _tokenize_ast(tree.function)
            yield Token(_LBRACKET, '[')
            yield from _tokenize_list(tree.arguments)
            yield Token(_RBRACKET, ']')
        case 'FunctionChange':
            yield Token('EXCLAMATION', '!')
            yield from _tokenize_ast(tree.item)
            yield Token('EQ', '=')
            yield from _tokenize_ast(tree.expression)
        case 'Have':
            yield _make_token('HAVE')
            yield from _tokenize_ast(
                tree.expression)
        case 'Hide':
            yield _make_token('HIDE')
            yield from _tokenize_comma_list(
                tree.facts)
            if tree.names:
                yield _make_token('DEF')
            yield from _tokenize_names(tree.names)
        case 'If':
            yield _make_token('IF')
            yield from _tokenize_ast(tree.predicate)
            yield _make_token('THEN')
            yield from _tokenize_ast(tree.then)
            yield _make_token('ELSE')
            yield from _tokenize_ast(tree.else_)
        case 'Instance':
            yield _make_token('INSTANCE')
            yield _make_name_token(tree.name)
            if tree.with_substitution is None:
                return
            yield _make_token('WITH')
            yield from _tokenize_list(
                tree.with_substitution)
        case 'VerticalList':
            for vertical_item in tree.arguments:
                yield from _tokenize_ast(
                    vertical_item)
        case 'VerticalItem':
            match tree.operator:
                case '/\\':
                    yield Token('AND', '/\\')
                case r'\/':
                    yield Token('OR', r'\/')
                case '|':
                    yield Token('PIPE', '|')
                case _:
                    raise ValueError(
                        tree.operator)
            yield from _tokenize_ast(
                tree.expression)
        case 'KeyBound':
            yield _make_name_token(tree.name)
            yield Token('COLON', ':')
            yield from _tokenize_ast(tree.bound)
        case 'Lambda':
            yield _make_token('LAMBDA')
            yield from _tokenize_comma_list(
                tree.parameters)
            yield Token('COLON', ':')
            yield from _tokenize_ast(
                tree.expression)
        case 'Let':
            yield _make_token('LET')
            yield from _tokenize_list(
                tree.definitions)
            yield _make_token('IN')
            yield from _tokenize_ast(
                tree.expression)
        case 'MapsTo':
            yield _make_name_token(tree.name)
            yield Token('MAPSTO', '|->')
            yield from _tokenize_ast(expression)
        case 'Module':
            yield Token(_DASH_LINE, _MIN_LINE_LEN * '-')
            yield Token('MODULE', 'MODULE')
            yield _make_name_token(tree.name)
            yield Token(_DASH_LINE, _MIN_LINE_LEN * '-')
            if tree.extendees:
                yield _make_token('EXTENDS')
            yield from _tokenize_comma_list(
                tree.extendees)
            for unit in tree.units:
                is_dash_line = (
                    isinstance(unit, str) and
                    unit.startswith('----'))
                if is_dash_line:
                    yield Token('DASH_LINE', unit)
                    continue
                yield from _tokenize_ast(unit)
            yield Token(_EQ_LINE, _MIN_LINE_LEN * '=')
        case 'ModuleName':
            yield Token('MODULE')
            yield _make_name_token(tree.name)
        case 'IntegralNumeral':
            # TODO: match prefix, and select
            # between decimal, binary, octal, hex
            yield Token('DECIMAL_INTEGER', tree.value)
        case 'Obvious':
            yield _make_token('OBVIOUS')
        case 'Omitted':
            yield _make_token('OMITTED')
        case 'OperatorApplication' if (
                tree.arguments is None):
            yield _make_name_token(tree.operator)
        case 'OperatorApplication':
            if isinstance(tree.operator, str):
                token_symbol = _lexemes_to_tokens.get(
                    tree.operator,
                    'NAME')
                token = Token(
                    symbol=token_symbol,
                    value=tree.operator)
                itr = tree.arguments
            else:
                itr = _tokenize_ast(tree.operator)
                token = next(itr)
            # TODO: between-operator `\X`
            if (token.value in
                    _TLA_OPERATOR_FIXITY['prefix'] or
                    token.value in
                        _TLA_OPERATOR_FIXITY['before']):
                # prefix
                arg, = tree.arguments
                yield token
                yield from _tokenize_ast(arg)
            # TODO: add all operators in this area
            elif (token.value in
                    _TLA_OPERATOR_FIXITY['infix'] or
                    token.value in
                        _TLA_OPERATOR_FIXITY['left'] or
                    token.value in
                        _TLA_OPERATOR_FIXITY['between']):
                # infix
                first, second = tree.arguments
                yield from _tokenize_ast(first)
                yield token
                yield from _tokenize_ast(second)
            elif (token.value in
                    _TLA_OPERATOR_FIXITY['postfix'] or
                    token.value in
                        _TLA_OPERATOR_FIXITY['after']):
                # postfix
                arg, = tree.arguments
                yield from _tokenize_ast(arg)
                yield token
            else:
                # nonfix
                yield token
                yield Token(_LPAREN, '(')
                yield from _tokenize_comma_list(itr)
                yield Token(_RPAREN, ')')
        case 'OperatorDeclaration':
            match tree.level:
                case None:
                    if not tree.new_keyword:
                        raise ValueError(tree)
                    level = None
                case (
                        'VARIABLE' |
                        'CONSTANT'|
                        'STATE' |
                        'ACTION' |
                        'TEMPORAL'):
                    level = tree.level
                case _:
                    raise ValueError(tree.level)
            if tree.new_keyword:
                yield _make_token('NEW')
            if level:
                yield _make_token(level)
            yield _make_name_token(tree.name)
            yield from _tokenize_arity(tree.arity)
            yield from _tokenize_bound(tree.bound)
        case 'OperatorDefinition':
            if tree.local:
                yield _make_token('LOCAL')
            yield _make_name_token(tree.name)
            if tree.arity is not None:
                yield Token(_LPAREN, '(')
                print(tree.arity)
                yield from _tokenize_comma_list(
                    tree.arity)
                yield Token(_RPAREN, ')')
            yield Token('DOUBLE_EQ', '==')
            yield from _tokenize_ast(tree.definiens)
        case 'OperatorName':
            raise NotImplementedError(
                'unused AST node type')
        case 'ParameterDeclaration':
            yield _make_name_token(tree.name)
            if tree.arity is None:
                return
            yield from _tokenize_ast(tree.arity)
        case 'Parentheses':
            yield Token(_LPAREN, '(')
            yield from _tokenize_ast(tree.expression)
            yield Token(_RPAREN, ')')
        case 'Pick':
            yield _make_token('PICK')
            yield from _tokenize_ast(tree.declarations)
            yield Token('COLON', ':')
            yield from _tokenize_ast(tree.predicate)
        case 'Proof':
            yield from _tokenize_list(tree.steps)
        case 'ProofCase':
            yield _make_token('CASE')
            yield from _tokenize_ast(tree.expression)
        case 'ProofStep':
            yield from _tokenize_ast(tree.name)
            yield from _tokenize_ast(tree.main)
            if tree.proof is None:
                return
            yield from _tokenize_ast(tree.proof)
        case 'Quantification':
            match tree.quantifier:
                case r'\A':
                    yield Token('BSLASH_A', r'\A')
                case r'\E':
                    yield Token('BSLASH_E', r'\E')
                case _:
                    raise ValueError(tree.quantifier)
            yield from _tokenize_comma_list(
                tree.declarations)
            yield Token('COLON', ':')
            yield from _tokenize_ast(tree.predicate)
        case 'QuantifierWitness':
            yield _make_token('WITNESS')
            yield from _tokenize_comma_list(
                tree.expressions)
        case 'Record':
            yield Token(_LBRACKET, '[')
            yield from _tokenize_ast(tree.key_values)
            yield Token(_RBRACKET, ']')
        case 'RecursiveOperators':
            yield _make_token('RECURSIVE')
            yield from _tokenize_names(tree.names)
        case 'Sequent':
            yield _make_token('ASSUME')
            yield from _tokenize_comma_list(
                tree.assume)
            yield _make_token('PROVE')
            yield from _tokenize_ast(tree.prove)
        case 'SetComprehension':
            yield Token(_LBRACE, '{')
            yield from _tokenize_ast(tree.item)
            yield Token('COLON', ':')
            yield from _tokenize_comma_list(
                tree.declarations)
            yield Token(_RBRACE, '}')
        case 'SetEnumeration':
            yield Token(_LBRACE, '{')
            yield from _tokenize_comma_list(tree.items)
            yield Token(_RBRACE, '}')
        case 'SetOfBoolean':
            yield _make_token('BOOLEAN')
        case 'SetOfFunctions':
            yield Token(_LBRACKET, '[')
            yield from _tokenize_ast(tree.domain)
            yield Token('RARROW', '->')
            yield from _tokenize_ast(tree.codomain)
            yield Token(_RBRACKET, ']')
        case 'SetOfRecords':
            yield Token(_LBRACKET, '[')
            yield from _tokenize_comma_list(
                tree.key_bounds)
            yield Token(_RBRACKET, ']')
        case 'SetOfStrings':
            yield _make_token('STRING')
        case 'SetSlice':
            yield Token(_LBRACE)
            for token in _tokenize_ast(tree.declaration):
                if token.symbol in {
                        'VARIABLE',
                        'CONSTANT', 'STATE', 'ACTION',
                        'TEMPORAL'}:
                    continue
                yield token
            yield Token('COLON', ':')
            yield from _tokenize_ast(tree.predicate)
            yield Token(_RBRACE)
        case 'StringLiteral':
            yield Token('STRING_LITERAL', tree.value)
        case 'SUBEXPRESSION_REFERENCE':
            for item in tree.items:
                yield from _tokenize_ast(item)
                yield Token('EXCLAMATION')
        case 'SubscriptedAction':
            match tree.operator:
                case '[':
                    opening = Token(_LBRACKET, '[')
                    closing = Token(
                        _RBRACKET_SUBSCRIPT, ']_')
                case '<<':
                    opening = Token(_LANGLE, '<<')
                    closing = Token(
                        _RANGLE_SUBSCRIPT, '>>_')
                case _:
                    raise ValueError(
                        tree.operator)
            yield opening
            yield from _tokenize_ast(tree.action)
            yield closing
            yield from _tokenize_ast(tree.subscript)
        case 'Suffices':
            yield _make_token('SUFFICES')
            yield from _tokenize_ast(tree.sequent)
        case 'Take':
            yield _make_token('TAKE')
            yield from _tokenize_comma_list(
                tree.declarations)
        case 'TemporalQuantification':
            match tree.quantifier:
                case '\\AA':
                    yield Token(
                        'SLASH_AA', '\\AA')
                case '\\EE':
                    yield Token(
                        'SLASH_EE', '\\EE')
                case _:
                    raise ValueError(
                        tree.quantifier)
            yield from _tokenize_names(
                tree.declarations)
            yield Token('COLON', ':')
            yield from _tokenize_ast(
                tree.predicate)
        case 'Theorem':
            yield _make_token('THEOREM')
            if tree.name:
                yield _make_name_token(tree.name)
                yield Token('DOUBLE_EQ', '==')
            yield from _tokenize_ast(tree.goal)
            if not tree.proof:
                return
            yield from _tokenize_ast(tree.proof)
        case 'Tuple':
            yield Token('DOUBLE_LANGLE', '<<')
            yield from _tokenize_comma_list(tree.items)
            yield Token('DOUBLE_RANGLE', '>>')
        case 'Use':
            yield _make_token('USE')
            if tree.only:
                yield _make_token('ONLY')
            yield from _tokenize_comma_list(tree.facts)
            if tree.names:
                yield _make_token('DEF')
            yield from _tokenize_names(tree.names)
        case 'Variables':
            yield _make_token('VARIABLE')
            yield from _tokenize_names(tree.names)
        case 'WithInstantiation':
            yield _make_name_token(tree.name)
            yield Token('LARROW', '<-')
            yield from _tokenize_ast(tree.expression)
        case _:
            raise ValueError(tree.symbol.name)


def _tokenize_bound(
        bound):
    if bound is None:
        return
    yield Token('BSLASH_IN', r'\in')
    yield from _tokenize_ast(bound)


def _tokenize_arity(
        arity):
    if arity is None:
        return
    if arity <= 0:
        raise ValueError(arity)
    yield Token(LPAREN, '(')
    yield from _tokenize_comma_list(
        '_' for _ in range(arity))
    yield Token(RPAREN, ')')


def _tokenize_list(
        container):
    if container is None:
        return
    for item in container:
        yield from _tokenize_ast(item)


def _tokenize_comma_list(
        iterable):
    if iterable is None:
        return
    first = True
    for item in iterable:
        if first:
            first = False
        else:
            yield Token('COMMA', ',')
        yield from _tokenize_ast(item)


def _tokenize_names(
        names):
    if names is None:
        return
    first = True
    for name in names:
        if first:
            first = False
        else:
            yield Token('COMMA', ',')
        match name:
            case str():
                yield _make_name_token(name)
            case _:
                yield from _tokenize_ast(name)


def _index_by_dfs(
        tree:
            tuple
        ) -> tuple:
    """Assign index attrbutes to nodes.

    The index value is the position of
    the node in a depth-first preorder
    traversal.

    If `tree` is an abstract syntax tree,
    then the indexing created by this
    function includes also nodes that
    are leaf attributes of namedtuples,
    strings, or lists.

    This is not an issue, because
    the indexing is unique, and this
    is the property that is used
    for mapping lexemes to AST nodes.
    """
    new_tree, _ = _index_by_dfs_recurse(
        tree,
        idx=0)
    return new_tree


def _index_by_dfs_recurse(
        tree,
        idx):
    node_index = idx
    idx += 1
    if isinstance(tree, tuple):
        successors = list()
        for succ in tree:
            new_succ, idx = _index_by_dfs_recurse(
                succ, idx)
            successors.append(new_succ)
        if hasattr(tree, 'node_index'):
            successors[-1] = node_index
            new_tree = type(tree)(*successors)
        else:
            new_tree = type(tree)(*successors)
    elif isinstance(tree, list):
        new_tree = list()
        for item in tree:
            new_item, idx = _index_by_dfs_recurse(
                item, idx)
            new_tree.append(new_item)
    else:
        new_tree = tree
    return new_tree, idx


def pprint_tree(
        tree):
    """Prettyprint syntax `tree`."""
    lines = pformat_tree(tree)
    print('\n'.join(lines))
    # _dump_tree_as_graph(tree)


def pformat_tree(
        tree):
    """Iterator of lines showing `tree`."""
    LINE = '\u2500' * 2
    BOTTOM_LEFT_CORNER = '\u2514'
    BRANCH = '\u251c'
    VERTICAL_LINE = '\u2502'
    END = f'{BOTTOM_LEFT_CORNER}{LINE}'
    SPACE = '\x20'
    if (isinstance(tree, tuple) and
            hasattr(tree, '_fields')):
        treed = tree._asdict()
    elif isinstance(tree, (list, tuple)):
        treed = {i: v for i, v in enumerate(tree)}
    else:
        yield f'{END} {tree}'
        return
    for i, u in enumerate(treed):
        prefix = (
            f'{BRANCH}{LINE}' if
                i < len(treed) - 1 else END)
        yield f'{prefix} {u}'
        if treed[u] is None:
            continue
        prefix = (VERTICAL_LINE if
            i < len(treed) - 1 else SPACE)
        for line in pformat_tree(treed[u]):
            yield f'{prefix}   {line}'



"""TLA+ context-free grammar with delimiters.

Definition also of operator precedence and associativity.
"""
import itertools as _itr
import typing as _ty

import tla._ast as _ast
import tla._utils as _utils


# The set of fixal operators
# that appear directly in the grammar.
EXCEPT_TOKEN_TYPES: _ty.Final = {
    'EQ', 'BSLASH_IN',
    'CARTESIAN', 'DASH',
    }
_utils.assert_len(
    EXCEPT_TOKEN_TYPES, 4)


class Grammar:
    """LR(1) grammar for parsing TLA+."""

    def __init__(
            self
            ) -> None:
        # self._make_operator_precedence_grammar()
        _utils.format_bnf_docstrings(self)
        _utils.check_bnf_grammar(self)
        self.tokens = None  # set by `Parser._build()`
        self.start = 'start'
        self.root_symbol = 'start'
        self._nodes = _ast.make_nodes()

    # NOTE: This method is unused, available for
    # building a context-free grammar for operators,
    # except for a few cases as for example `A # ~ B`.
    # These cases are handled by the
    # operator-precedence parser that results from
    # using an operator precedence table to
    # resolve shift-reduce ambiguities.
    def _make_operator_precedence_grammar(
            self
            ) -> None:
        """Add equalities about operators.

        These grammar equalities describe the
        precedence, associativity, fixity, and
        arity of prefix, infix, and postfix
        operators.
        """
        # Associativity and precedence of
        # prefix, infix, and postfix operators.
        # This tuple is used to generate the
        # part of the grammar that describes
        # the application of these operators.
        positioning = (
            'OP_1_INFIX',
            'OP_2_INFIX',
            'OP_3_LEFT',
            'OP_4_PREFIX',
            'OP_4_BEFORE',
            'op_5_eq_infix',
            'op_5_bslash_in_between',
            'OP_5_LEFT',
            'OP_5_INFIX',
            'OP_6_LEFT',
            'OP_7_INFIX',
            'OP_8_BEFORE',
            'OP_8_INFIX',
            'OP_8_LEFT',
            'OP_9_BEFORE',
            'OP_9_INFIX',
            'OP_9_LEFT',
            'OP_10_LEFT',
            'OP_10_INFIX',
            'set_of_tuples_between',
            'OP_11_LEFT',
            'op_11_dash_left',
            'op_12_dash_before',
            'OP_12_PREFIX',
                # for `DASH_DOT`
            'OP_13_INFIX',
            'OP_13_LEFT',
            'OP_14_INFIX',
            'OP_15_POSTFIX',
            'OP_15_AFTER',
            # the last name is not mapped
            # to a method automatically
            'function_application',
            )
        tokens_of_symbols = dict(
            op_5_eq_infix='EQ',
            op_5_bslash_in_between='BSLASH_IN',
            op_11_dash_left='DASH',
            op_12_dash_before='DASH',
            set_of_tuples_between='CARTESIAN',
            )
        _utils.make_operator_precedence_grammar(
            positioning, tokens_of_symbols, self)

    def __str__(
            self
            ) -> str:
        """Return grammar of TLA+.

        The grammar originates in
        method docstrings.
        """
        (_, _, _, grammar_text
            ) = _utils.collect_grammar_info(
                self,
                add_braces=False)
        return grammar_text.replace(
            '/\\', '\x20\x20', 1)

    def p_error(
            self,
            token):
        """Handle errors."""
        # tokens = list()
        # while True:
        #     token = self._parser.token()
        #     if token is None:
        #         break
        #     tokens.append(token)
        # rest = ' '.join(tokens)
        raise ValueError(
            f'Syntax error at `{token!r}`\n')

    def p_start(
            self, p):
        """start: tla_module"""
        p[0] = p[1]

    def p_module_that_extends(
            self, p):
        """tla_module:
            module_signature
            extends
            units
            EQ_LINE
        """
        p[0] = self._nodes.Module(
            name=p[1],
            extendees=p[2],
            units=p[3])

    def p_module_without_extends(
            self, p):
        """tla_module:
            module_signature
            units
            EQ_LINE
        """
        p[0] = self._nodes.Module(
            name=p[1],
            units=p[2])

    def p_module_with_extends_only(
            self, p):
        """tla_module:
            module_signature
            extends
            EQ_LINE
        """
        p[0] = self._nodes.Module(
            name=p[1],
            units=p[2])

    def p_empty_module(
            self, p):
        """tla_module:
            module_signature
            EQ_LINE
        """
        p[0] = self._nodes.Module(
            name=p[1])

    def p_module_signature(
            self, p):
        """module_signature:
            DASH_LINE expr DASH_LINE
        """
        p[0] = p[2].name

    def p_module_name(
            self, p):
        """expr: MODULE NAME"""
        # This allows `MODULE NAME` to
        # appear inside `DEF` sections
        p[0] = self._nodes.ModuleName(
            name=p[2])

    def p_extends_statement(
            self, p):
        """extends:
            EXTENDS exprs
        """
        p[0] = p[2]

    def p_units_append(
            self, p):
        """units: units unit"""
        p[1].append(p[2])
        p[0] = p[1]

    def p_units_start(
            self, p):
        """units: unit"""
        p[0] = [p[1]]

    def p_unit(
            self, p):
        """unit:
            | operator_def
            | variable_decl
            | constant_decl
            | recursive_decl
            | use_or_hide
            | theorem
            | axiom
            | tla_module
            | instance
            | DASH_LINE
        """
        p[0] = p[1]

    def p_variable_declaration(
            self, p):
        """variable_decl:
            variable_keyword exprs
        """
        p[0] = self._nodes.Variables(
            names=p[2])

    def p_constant_declaration(
            self, p):
        """constant_decl:
            constant_keyword exprs
        """
        p[0] = self._nodes.Constants(
            names=p[2])

    def p_recursive_declaration(
            self, p):
        """recursive_decl:
            RECURSIVE exprs
        """
        p[0] = self._nodes.RecursiveOperators(
            names=p[2])

    def p_keyword_variable(
            self, p):
        """variable_keyword:
            | VARIABLES
            | VARIABLE
        """
        p[0] = p[1]

    def p_keyword_constant(
            self, p):
        """constant_keyword:
            | CONSTANTS
            | CONSTANT
        """
        p[0] = p[1]

    def p_operator_definitions_append(
            self, p):
        """operator_defs:
            operator_defs operator_def
        """
        p[1].append(p[2])
        p[0] = p[1]

    def p_operator_definitions_start(
            self, p):
        """operator_defs: operator_def"""
        p[0] = [p[1]]

    def p_nonlocal_operator_definition(
            self, p):
        """operator_def:
            DEF_SIG
                expr DOUBLE_EQ
                    definiens
        """
        # NOTE: `expr` here for function definitions:
        # `f[x \in R, y \in S] == ...`
        sig = p[2]
        definiens = p[4]
        p[0] = self._make_operator_definition(
            sig, definiens)

    def _make_operator_definition(
            self,
            sig,
            definiens,
            local=None):
        """Return operator definition."""
        if sig.symbol.name == 'FUNCTION_APPLICATION':
            # function definition
            func_sig = sig.function
            if func_sig.arguments is not None:
                raise ValueError(sig)
            name = func_sig.operator
            exprs = sig.arguments
            definiens = self._nodes.Function(
                declaration=exprs,
                value=definiens)
                # NOTE: causes printing to differ
                # from parsed text
            arity = None
            function = True
        else:
            # operator definition
            name = sig.operator
            arity = sig.arguments
            function = None
        return self._nodes.OperatorDefinition(
            name=name,
            arity=arity,
            definiens=definiens,
            local=local,
            function=function)

    def p_local_operator_definition(
            self, p):
        """operator_def:
            DEF_SIG LOCAL
                expr DOUBLE_EQ
                    definiens
        """
        sig = p[3]
        definiens = p[5]
        local = True
        p[0] = self._make_operator_definition(
            sig, definiens, local)

    # def p_operator_arity_without_args(
    #         self, p):
    #     """operator_arity:
    #         NAME
    #     """
    #     p[0] = self._nodes.OperatorApplication(
    #         operator=p[1])
    #
    # def p_operator_arity_with_args(
    #         self, p):
    #     """operator_arity:
    #         NAME LPAREN
    #             operator_arity_args
    #         RPAREN
    #     """
    #     # NOTE: pattern for higher-order operators
    #     p[0] = self._nodes.OperatorApplication(
    #         operator=p[1],
    #         arguments=p[3])
    #
    # def p_operator_arity_args_append(
    #         self, p):
    #     """operator_arity_args:
    #         operator_arity_args COMMA
    #         operator_arity_arg
    #     """
    #     p[0] = p[1]
    #     p[1].append(p[3])
    #
    # def p_operator_arity_args_start(
    #         self, p):
    #     """operator_arity_args:
    #         operator_arity_arg
    #     """
    #     p[0] = [p[1]]
    #
    # def p_operator_arity_arg(
    #         self, p):
    #     """operator_arity_arg:
    #         | operator_arity
    #         | UNDERSCORE
    #     """
    #     p[0] = p[1]

    def p_definiens(
            self, p):
        """definiens:
            | expr
            | instance
        """
        p[0] = p[1]

    def p_use_or_hide(
            self, p):
        """use_or_hide:
            | use_unit
            | hide_unit
        """
        p[0] = p[1]

    def p_use_unit_full(
            self, p):
        """use_unit:
            use_only
                exprs
                def_keyword name_list
        """
        p[0] = self._nodes.Use(
            only=p[1],
            facts=p[2],
            names=p[4])

    def p_use_unit_with_facts(
            self, p):
        """use_unit:
            use_only
                exprs
        """
        p[0] = self._nodes.Use(
            only=p[1],
            facts=p[2])

    def p_use_unit_with_defs(
            self, p):
        """use_unit:
            use_only
                def_keyword name_list
        """
        p[0] = self._nodes.Use(
            only=p[1],
            names=p[3])

    def p_module_unit_hide(
            self, p):
        """hide_unit:
            HIDE
                exprs
                def_keyword name_list
        """
        p[0] = self._nodes.Hide(
            facts=p[2],
            names=p[4])

    def p_use_without_only(
            self, p):
        """use_only: USE"""
        p[0] = False

    def p_use_with_only(
            self, p):
        """use_only: USE ONLY"""
        p[0] = True

    def p_name_list_append(
            self, p):
        """name_list:
            | name_list COMMA general_identifier
            | name_list COMMA MODULE NAME
            | name_list COMMA
                VERTICAL_OPERATOR
                LVERTICAL
        """
        if p[3] == 'MODULE':
            name = self._nodes.ModuleName(
                name=p[4])
        else:
            name = p[3]
        p[1].append(name)
        p[0] = p[1]

    def p_name_list_start(
            self, p):
        """name_list:
            | general_identifier
            | MODULE NAME
            | VERTICAL_OPERATOR
              LVERTICAL
        """
        if p[1] == 'MODULE':
            name = self._nodes.ModuleName(
                name=p[2])
        else:
            name = p[1]
        p[0] = [name]

    def p_named_axiom(
            self, p):
        """axiom:
            axiom_keyword
                DEF_SIG NAME DOUBLE_EQ
                    expr
        """
        p[0] = self._nodes.Axiom(
            name=p[3],
            expression=p[5],
            axiom_keyword=p[1])

    def p_unnamed_axiom(
            self, p):
        """axiom: axiom_keyword expr"""
        p[0] = self._nodes.Axiom(
            expression=p[2],
            axiom_keyword=p[1])

    def p_axiom_keyword(
            self, p):
        """axiom_keyword:
            | AXIOM
            | ASSUME
            | ASSUMPTION
        """
        p[0] = p[1]

    def p_named_theorem(
            self, p):
        """theorem:
            | theorem_keyword
                DEF_SIG NAME DOUBLE_EQ
                    sequent
                    theorem_proof
            | theorem_keyword
                DEF_SIG NAME DOUBLE_EQ
                    expr
                    theorem_proof
        """
        p[0] = self._nodes.Theorem(
            theorem_keyword=p[1],
            name=p[3],
            goal=p[5],
            proof=p[6])

    def p_unnamed_theorem(
            self, p):
        """theorem:
            | theorem_keyword
                sequent
                theorem_proof
            | theorem_keyword
                expr
                theorem_proof
        """
        p[0] = self._nodes.Theorem(
            theorem_keyword=p[1],
            goal=p[2],
            proof=p[3])

    def p_named_theorem_without_proof(
            self, p):
        """theorem:
            | theorem_keyword
                DEF_SIG NAME DOUBLE_EQ
                    sequent
            | theorem_keyword
                DEF_SIG NAME DOUBLE_EQ
                    expr
        """
        p[0] = self._nodes.Theorem(
            theorem_keyword=p[1],
            name=p[3],
            goal=p[5])

    def p_unnamed_theorem_without_proof(
            self, p):
        """theorem:
            | theorem_keyword
                sequent
            | theorem_keyword
                expr
        """
        p[0] = self._nodes.Theorem(
            theorem_keyword=p[1],
            goal=p[2])

    def p_theorem_keyword(
            self, p):
        """theorem_keyword:
            | PROPOSITION
            | LEMMA
            | THEOREM
            | COROLLARY
        """
        p[0] = p[1]

    def p_sequent(
            self, p):
        """sequent:
            ASSUME asms
            PROVE expr
        """
        p[0] = self._nodes.Sequent(
            assume=p[2],
            prove=p[4])

    def p_assumptions_append(
            self, p):
        """asms: asms COMMA asm"""
        p[1].append(p[3])
        p[0] = p[1]

    def p_assumptions_start(
            self, p):
        """asms: asm"""
        p[0] = [p[1]]

    def p_assumption_new_decl(
            self, p):
        """asm: NEW asm_decl"""
        decl = p[2]
        p[0] = self._nodes.OperatorDeclaration(
            name=decl.name,
            level=decl.level,
            new_keyword=True)

    def p_assumption(
            self, p):
        """asm: asm_decl"""
        p[0] = p[1]

    def p_assumption_new_constant(
            self, p):
        """asm: NEW expr"""
        # read `tla._ast.op_decl_name_bound_arity()`
        #
        # TODO: check that this is one of:
        # - unbounded NAME
        # - bounded NAME
        # - unbounded non-nullary
        #   nonfixal declaration
        # - unbounded fixal declaration
        p[0] = self._nodes.OperatorDeclaration(
            name=p[2],
            new_keyword=True)

    def p_assumption_sequent(
            self, p):
        """asm:
            | sequent
            | expr
        """
        p[0] = p[1]

    def p_variable_declaration_in_sequent(
            self, p):
        """asm_decl: VARIABLE NAME"""
        p[0] = self._nodes.OperatorDeclaration(
            name=p[2],
            level='VARIABLE')

    def p_unbounded_constant_decl_in_sequent(
            self, p):
        """asm_decl: CONSTANT expr"""
        # method `tla._ast.op_decl_name_bound_arity()`
        # gets the name, bound, and arity from `expr`
        #
        # TODO: check that this is either:
        # - bounded nullary
        # - unbounded nullary
        # - unbounded non-nullary
        #   nonfixal declaration
        # - unbounded fixal declaration
        p[0] = self._nodes.OperatorDeclaration(
            name=p[2],
            level=p[1])

    def p_levely_declaration_in_sequent(
            self, p):
        """asm_decl: level_specifier expr"""
        # read `tla._ast.op_decl_name_bound_arity()`
        p[0] = self._nodes.OperatorDeclaration(
            name=p[2],
            level=p[1])

    def p_state_level_specifier(
            self, p):
        """level_specifier: STATE"""
        # The keyword `CONSTANT`, too,
        # is a level specifier.
        # The parser productions for
        # `CONSTANT` declarations in sequents
        # are in another area of the grammar
        # in order to avoid
        # shift/reduce ambiguities when
        # building the transitions table of
        # the LALR(1) parser.
        p[0] = 'STATE_LEVEL'

    def p_action_level_specifier(
            self, p):
        """level_specifier: ACTION"""
        p[0] = 'ACTION_LEVEL'

    def p_temporal_level_specifier(
            self, p):
        """level_specifier: TEMPORAL"""
        p[0] = 'TEMPORAL_LEVEL'

    def p_parentheses(
            self, p):
        """paren: LPAREN expr RPAREN"""
        p[0] = self._nodes.Parentheses(
            expression=p[2],)

    def p_instance_verbatim_local(
            self, p):
        """instance: LOCAL INSTANCE NAME"""
        p[0] = self._nodes.Instance(
            name=p[3],
            local=True)

    def p_instance_verbatim_nonlocal(
            self, p):
        """instance: INSTANCE NAME"""
        p[0] = self._nodes.Instance(
            name=p[2])

    def p_instance_with_substitution(
            self, p):
        """instance:
            INSTANCE NAME WITH substitutions
        """
        p[0] = self._nodes.Instance(
            name=p[2],
            with_substitution=p[4])

    def p_substitutions_append(
            self, p):
        """substitutions:
            substitutions COMMA sub
        """
        p[1].append(p[3])
        p[0] = p[1]

    def p_substitutions_start(
            self, p):
        """substitutions: sub"""
        p[0] = [p[1]]

    def p_substitution(
            self, p):
        """sub:
            | replaced_expr LARROW expr
            | replaced_expr LARROW
                vertical_op LVERTICAL RVERTICAL
        """
        name = p[1]
        if not isinstance(name, str):
            raise AssertionError(name)
        p[0] = self._nodes.WithInstantiation(
            name=name,
            expression=p[3])

    def p_replaced_expr(
            self, p):
        """replaced_expr :
            | NAME
            | OP_1_INFIX
            | OP_2_INFIX
            | OP_3_LEFT
            | OP_4_PREFIX
            | OP_4_BEFORE
            | EQ
            | BSLASH_IN
            | OP_5_LEFT
            | OP_5_INFIX
            | OP_6_LEFT
            | OP_7_INFIX
            | OP_8_BEFORE
            | OP_8_INFIX
            | OP_8_LEFT
            | OP_9_BEFORE
            | OP_9_INFIX
            | OP_9_LEFT
            | OP_10_LEFT
            | OP_10_INFIX
            | CARTESIAN
            | OP_11_LEFT
            | DASH
            | OP_12_PREFIX
            | OP_13_INFIX
            | OP_13_LEFT
            | OP_14_INFIX
            | OP_15_POSTFIX
            | OP_15_AFTER
            | VERTICAL_OPERATOR LVERTICAL
        """
        p[0] = p[1]

    def p_lambda(
            self, p):
        """expr:
            LAMBDA exprs COLON
                expr END_LAMBDA
        """
        parameters = p[2]
        for param in parameters:
            if param.arguments:
                raise ValueError(
                    'LAMBDA parameter: ',
                    param)
        p[0] = self._nodes.Lambda(
            parameters=parameters,
            expression=p[4])

    def p_undelimited(
            self, p):
        """undelimited_expr:
            | temporal_quantification
            | constant_quantification
            | choose_expr
            | let_expr
            | case_expr
            | ternary_expr
            | elif_expr
        """
        p[0] = p[1]

    def p_temporal_quantification(
            self, p):
        """temporal_quantification:
            temporal_quantifier exprs COLON
                expr
            end_temporal_quantifier
        """
        p[0] = self._nodes.TemporalQuantification(
            quantifier=p[1],
            declarations=p[2],
            predicate=p[4])

    def p_constant_quantification(
            self, p):
        """constant_quantification:
            quantifier exprs COLON
                expr
            end_quantifier
        """
        p[0] = self._nodes.Quantification(
            quantifier=p[1],
            declarations=p[2],
            predicate=p[4])

    def p_choose(
            self, p):
        """choose_expr:
            CHOOSE expr COLON expr
            END_CHOOSE
        """
        # The grammar symbol `expr` is used
        # here, and analysis after parsing
        # will check that the items between
        # the angles are names or a tuple.
        p[0] = self._nodes.Choose(
            declaration=p[2],
            predicate=p[4])

    def p_let_in(
            self, p):
        """let_expr:
            LET operator_defs
            IN expr
            END_LET
        """
        p[0] = self._nodes.Let(
            definitions=p[2],
            expression=p[4])

    def p_case_expr_without_default(
            self, p):
        """case_expr:
            CASE cases END_CASE
        """
        p[0] = self._nodes.Cases(
            cases=p[2])

    def p_case_expr_with_other(
            self, p):
        """case_expr:
            CASE cases other_case END_CASE
        """
        p[0] = self._nodes.Cases(
            cases=p[2],
            other=p[3])

    def p_cases_append(
            self, p):
        """cases:
            cases
            OP_4_BEFORE case
        """
        if p[2] != '[]':
            raise ValueError(
                'syntax error in CASE expression')
        p[1].append(p[3])
        p[0] = p[1]

    def p_cases_start(
            self, p):
        """cases: case"""
        p[0] = [p[1]]

    def p_rarrow_infix_non(
            self, p):
        """case: expr RARROW expr"""
        p[0] = self._nodes.CaseItem(
            predicate=p[1],
            expression=p[3])

    def p_other_case(
            self, p):
        """other_case:
            OP_4_BEFORE
                OTHER RARROW expr
        """
        if p[1] != '[]':
            raise ValueError(
                'syntax error in CASE expression')
        p[0] = p[4]

    def p_ternary(
            self, p):
        """ternary_expr:
            IF expr
                THEN expr
                ELSE expr
                END_IF
        """
        p[0] = self._nodes.If(
            predicate=p[2],
            then=p[4],
            else_=p[6])

    def p_conditional(
            self, p):
        """elif_expr:
            IF expr
                THEN expr
                elifs
                ELSE expr
                END_IF
        """
        case = self._nodes.IfItem(
            predicate=p[2],
            expression=p[4])
        cases = [case] + p[5]
        p[0] = self._nodes.Ifs(
            cases=cases,
            other=p[7])

    def p_elifs_append(
            self, p):
        """elifs: elifs elif_"""
        p[1].append(p[2])
        p[0] = p[1]

    def p_elifs_start(
            self, p):
        """elifs: elif_"""
        p[0] = [p[1]]

    def p_elif_(
            self, p):
        """elif_: ELIF expr THEN expr"""
        p[0] = self._nodes.IfItem(
            predicate=p[2],
            expression=p[4])

    # def p_expr_op_1_infix_non(
    #         self, p):
    #     """expr: op_1_infix"""
    #     # The method `p_op_1_infix()` is
    #     # dynamically created by the
    #     # method `__init__()`.
    #     p[0] = p[1]

    def p_function_application(
            self, p):
        """function_application:
            expr
                LBRACKET exprs RBRACKET
        """
        # The grammar symbol
        # `function_application` is used
        # in the grammar rule of the grammar
        # symbol `op_15_postfix`.
        #
        # The grammar symbol `op_15_postfix`
        # appears in the docstring of the
        # method `p_op_15_postfix()`,
        # which is dynamically created by
        # the method `__init__()`.
        p[0] = self._nodes.FunctionApplication(
            function=p[1],
            arguments=p[3])

    def p_function_application_skip(
            self, p):
        """function_application:
            op_record_field
        """
        # The precedence of record fields,
        # relative to function application,
        # is defined on page 284 of
        # the book "Specifying Systems".
        p[0] = p[1]

    def p_record_field(
            self, p):
        """op_record_field: fields"""
        p[0] = p[1]

    def p_fields(
            self, p):
        """fields: expr DOT NAME"""
        p[0] = self._nodes.Field(
            expression=p[1],
            name=p[3])

    def p_fields_start(
            self, p):
        """fields:
            | delimited_expr
            | general_identifier
        """
        p[0] = p[1]

    def p_record_field_skip(
            self, p):
        """op_record_field:
            | undelimited_expr
            | vitems
            | subscripted_expr
        """
        p[0] = p[1]

    def p_vertical_items(
            self, p):
        """vitems: vertical_items"""
        vertical_items = p[1]
        if not vertical_items:
            raise AssertionError(
                'empty vertical list')
        vertical_item, *_ = vertical_items
        op = vertical_item.operator
        # that all vertical items have the
        # same operator is checked after parsing
        p[0] = self._nodes.VerticalList(
            operator=op,
            arguments=p[1])

    def p_vertical_items_append(
            self, p):
        """vertical_items:
            vertical_items vertical_item
        """
        p[1].append(p[2])
        p[0] = p[1]

    def p_vertical_items_start(
            self, p):
        """vertical_items:
            vertical_item
        """
        p[0] = [p[1]]

    def p_vertical_item(
            self, p):
        """vertical_item:
            vertical_op
                LVERTICAL
                expr
                RVERTICAL
        """
        p[0] = self._nodes.VerticalItem(
            operator=p[1],
            expression=p[3])

    def p_vertical_operator(
            self, p):
        """vertical_op: VERTICAL_OPERATOR"""
        p[0] = p[1]

    def p_temporal_quantifier(
            self, p):
        """temporal_quantifier:
            | BSLASH_AA
            | BSLASH_EE
        """
        p[0] = p[1]

    def p_end_temporal_quantifier(
            self, p):
        """end_temporal_quantifier:
            | END_BSLASH_AA
            | END_BSLASH_EE
        """

    def p_quantifier(
            self, p):
        """quantifier:
            | BSLASH_A
            | BSLASH_E
        """
        p[0] = p[1]

    def p_end_quantifier(
            self, p):
        """end_quantifier:
            | END_BSLASH_A
            | END_BSLASH_E
        """

    def p_operator_application(
            self, p):
        """op_app:
            | nullary_app
            | unary_app
            | multiary_app
        """
        p[0] = p[1]

    def p_unary_operator_application(
            self, p):
        """unary_app: NAME LPAREN expr RPAREN"""
        args = [p[3]]
        p[0] = self._nodes.OperatorApplication(
            operator=p[1],
            arguments=args)

    def p_multiary_operator_application(
            self, p):
        """multiary_app:
            NAME LPAREN exprs_2 RPAREN"""
        p[0] = self._nodes.OperatorApplication(
            operator=p[1],
            arguments=p[3])

    def p_nullary_operator_application(
            self, p):
        """nullary_app: NAME"""
        p[0] = self._nodes.OperatorApplication(
            operator=p[1])

    def p_subscripted_expr(
            self, p):
        """subscripted_expr:
            | subscripted_action
            | fairness
        """
        p[0] = p[1]

    def p_subscripted_action(
            self, p):
        """subscripted_action:
            | LBRACKET expr
                RBRACKET_UNDERSCORE
                    subscript
            | DOUBLE_LANGLE expr
                DOUBLE_RANGLE_UNDERSCORE
                    subscript
            | LBRACKET expr
                RBRACKET_UNDERSCORE
                    NAME
            | DOUBLE_LANGLE expr
                DOUBLE_RANGLE_UNDERSCORE
                    NAME
        """
        p[0] = self._nodes.SubscriptedAction(
            operator=p[1],
            action=p[2],
            subscript=p[4])

    def p_fairness(
            self, p):
        """fairness:
            | WF_ subscript
                LPAREN expr RPAREN
            | SF_ subscript
                LPAREN expr RPAREN
            | WF_ NAME
                LPAREN expr RPAREN
            | SF_ NAME
                LPAREN expr RPAREN
        """
        subscript = p[2]
        if isinstance(subscript, str):
            subscript = self._nodes.OperatorApplication(
                operator=subscript)
        p[0] = self._nodes.Fairness(
            operator=p[1],
            subscript=subscript,
            action=p[4])

    def p_subscript_unary_op_application(
            self, p):
        """subscript: NAME LPAREN expr RPAREN"""
        args = [p[3]]
        p[0] = self._nodes.OperatorApplication(
            operator=p[1],
            arguments=args)

    def p_subscript_multiary_op_application(
            self, p):
        """subscript:
            NAME LPAREN exprs_2 RPAREN"""
        p[0] = self._nodes.OperatorApplication(
            operator=p[1],
            arguments=p[3])

    def p_subscript_delimited_expr(
            self, p):
        """subscript: delimited_expr"""
        p[0] = p[1]

    def p_delimited_expr(
            self, p):
        """delimited_expr:
            | set_enum
            | deferred_set
            | set_comprehension
            | set_slice
            | tuple
            | record
            | set_of_records
            | set_of_functions
            | value_map
            | paren
            | leaf
        """
        # | function_constructor
        p[0] = p[1]

    def p_empty_set_enumeration(
            self, p):
        """set_enum: LBRACE RBRACE"""
        p[0] = self._nodes.SetEnumeration(
            items=list())

    def p_set_enumeration(
            self, p):
        """set_enum: LBRACE exprs RBRACE"""
        p[0] = self._nodes.SetEnumeration(
            items=p[2])

    def p_deferred_set(
            self, p):
        """deferred_set:
            LBRACE expr COLON
                expr RBRACE
        """
        # Section 15.2.5 on page 289.
        first = p[2]
        second = p[4]
        op_1 = (
            hasattr(first, 'operator') and
            first.operator == '\\in')
        op_2 = (
            hasattr(second, 'operator') and
            second.operator == '\\in')
        match (op_1, op_2):
            case True, _:
                p[0] = self._nodes.SetSlice(
                    declaration=first,
                    predicate=second)
            case False, True:
                p[0] = self._nodes.SetComprehension(
                    item=first,
                    declarations=[second])
            case _:
                raise ValueError(
                    'At least one operator of '
                    '`{...: ...}` need be `\\in`. '
                    f'{first = }\n{second = }')

    def p_definitely_set_comprehension(
            self, p):
        """set_comprehension:
            LBRACE expr COLON
                exprs_2 RBRACE
        """
        p[0] = self._nodes.SetComprehension(
            item=p[2],
            declarations=p[4])

    def p_definitely_set_slice(
            self, p):
        """set_slice:
            LBRACE exprs_2 COLON
                expr RBRACE
        """
        raise ValueError(
            f'Syntax error {p[2]}')

    def p_set_of_functions(
            self, p):
        """set_of_functions:
            LBRACKET expr RARROW expr RBRACKET
        """
        p[0] = self._nodes.SetOfFunctions(
            domain=p[2],
            codomain=p[4])

    # def p_function_constructor(
    #         self, p):
    #     """function_constructor:
    #         LBRACKET exprs MAPSTO
    #             expr RBRACKET
    #     """
    #     p[0] = self._nodes.Function(
    #         declaration=p[2],
    #         value=p[4])

    def p_function_except(
            self, p):
        """value_map:
            LBRACKET expr EXCEPT
                modifications RBRACKET
        """
        p[0] = self._nodes.Except(
            function=p[2],
            changes=p[4])

    def p_modifications_append(
            self, p):
        """modifications:
            modifications COMMA modification
        """
        p[1].append(p[3])
        p[0] = p[1]

    def p_modifications_start(
            self, p):
        """modifications: modification"""
        p[0] = [p[1]]

    def p_modification(
            self, p):
        """modification:
            EXCLAMATION modificandum EQ expr
        """
        p[0] = self._nodes.FunctionChange(
            item=p[2],
            expression=p[4])

    def p_modificandum_append(
            self, p):
        """modificandum:
            modificandum func_value
        """
        p[1].append(p[2])
        p[0] = p[1]

    def p_modificandum_start(
            self, p):
        """modificandum: func_value"""
        p[0] = [p[1]]

    def p_except_bracket(
            self, p):
        """func_value: LBRACKET exprs RBRACKET"""
        p[0] = p[2]

    # NOTE: keywords too can occur here
    # `tla._preparser.switch_keywords_after_dot()`
    # changes keywords after `DOT` to `NAME`.
    def p_except_dot(
            self, p):
        """func_value: DOT NAME"""
        p[0] = p[2]

    def p_empty_tuple(
            self, p):
        """tuple: DOUBLE_LANGLE DOUBLE_RANGLE"""
        p[0] = self._nodes.Tuple(
            items=list())

    def p_tuple(
            self, p):
        """tuple: DOUBLE_LANGLE exprs DOUBLE_RANGLE"""
        p[0] = self._nodes.Tuple(
            items=p[2])

    def p_record(
            self, p):
        """record:
            LBRACKET mapsto_pairs RBRACKET
        """
        pairs = p[2]
        # function ?
        if len(pairs) == 1:
            decl = pairs[0].declaration
            is_func = (
                len(decl) > 1 or
                    # [x \in S, y \in R |-> ...]
                decl[0].arguments is not None)
                    # [x \in S |-> ...]
            if is_func:
                p[0] = pairs[0]
                return
        # record
        key_values = list()
        for pair in pairs:
            if len(pair.declaration) > 1:
                raise ValueError(t.declaration)
            decl = pair.declaration[0]
            if decl.arguments is not None:
                raise ValueError(decl)
            key = decl.operator
            value = pair.value
            key_value = (key, value)
            key_values.append(key_value)
        p[0] = self._nodes.Record(
            key_values=key_values)

    def p_set_of_records(
            self, p):
        """set_of_records:
            LBRACKET colon_pairs RBRACKET
        """
        p[0] = self._nodes.SetOfRecords(
            key_bounds=p[2])

    def p_mapsto_pairs_append(
            self, p):
        """mapsto_pairs:
            mapsto_pairs COMMA mapsto_pair
        """
        p[1].append(p[3])
        p[0] = p[1]

    def p_mapsto_pairs_start(
            self, p):
        """mapsto_pairs: mapsto_pair"""
        p[0] = [p[1]]

    def p_mapsto_pair(
            self, p):
        """mapsto_pair: exprs MAPSTO expr"""
        # NOTE: For converting to bounds:
        # `tla._ast.declarations_as_name_bounds()`
        # which takes
        # `x, y \in S` and returns
        # `[(x, S), `(y, S)]`
        p[0] = self._nodes.Function(
            declaration=p[1],
            value=p[3])

    def p_colon_pairs_append(
            self, p):
        """colon_pairs:
            colon_pairs COMMA colon_pair
        """
        p[1].append(p[3])
        p[0] = p[1]

    def p_colon_pairs_start(
            self, p):
        """colon_pairs: colon_pair"""
        p[0] = [p[1]]

    def p_colon_pair(
            self, p):
        """colon_pair: NAME COLON expr"""
        p[0] = self._nodes.KeyBound(
            name=p[1],
            bound=p[3])

    def p_exprs_1(
            self, p):
        """exprs: expr"""
        p[0] = [p[1]]

    def p_exprs_2(
            self, p):
        """exprs: exprs_2"""
        p[0] = p[1]

    def p_exprs_append(
            self, p):
        """exprs_2:
            exprs_2 COMMA expr
        """
        p[1].append(p[3])
        p[0] = p[1]

    def p_exprs_start(
            self, p):
        """exprs_2:
            expr COMMA expr
        """
        p[0] = [p[1], p[3]]

    def p_leaf(
            self, p):
        """leaf:
            | integer_literal
            | float_literal
            | string_literal
            | boolean_literal
            | set_of_booleans
            | set_of_strings
            | UNDERSCORE
        """
        p[0] = p[1]

    def p_integer_literal(
            self, p):
        """integer_literal:
            | DECIMAL_INTEGER
            | BINARY_INTEGER
            | OCTAL_INTEGER
            | HEXADECIMAL_INTEGER
        """
        # This is a numeral.
        # When a TLA+ module `EXTENDS`
        # `Naturals` or `Integers`,
        # then these numerals take
        # the usual integer meaning.
        p[0] = self._nodes.IntegralNumeral(
            value=p[1])

    def p_float_literal(
            self, p):
        """float_literal: FLOAT"""
        p[0] = self._nodes.FloatNumeral(
            value=p[1])

    def p_string_literal(
            self, p):
        """string_literal:
            | STRING_LITERAL
            | MULTILINE_STRING_LITERAL
        """
        p[0] = self._nodes.StringLiteral(
            value=p[1])

    def p_boolean_constant(
            self, p):
        """boolean_literal:
            | FALSE
            | TRUE
        """
        p[0] = self._nodes.BooleanLiteral(
            value=p[1])

    def p_set_of_boolean_constants(
            self, p):
        """set_of_booleans: BOOLEAN"""
        p[0] = self._nodes.SetOfBooleans()

    def p_set_of_strings(
            self, p):
        """set_of_strings: STRING"""
        p[0] = self._nodes.SetOfStrings()

    def p_theorem_proof_steps(
            self, p):
        """theorem_proof:
            PROOF proof_steps
        """
        p[0] = self._nodes.Proof(
            steps=p[2],
            proof_keyword=True)

    def p_theorem_leaf_proof(
            self, p):
        """theorem_proof:
            PROOF leaf_proof
        """
        p[0] = self._nodes.Proof(
            steps=[p[2]],
            proof_keyword=True)

    def p_theorem_proof_steps_without_keyword(
            self, p):
        """theorem_proof:
            proof_steps
        """
        p[0] = self._nodes.Proof(
            steps=p[1])

    def p_theorem_leaf_proof_without_keyword(
            self, p):
        """theorem_proof:
            leaf_proof
        """
        p[0] = self._nodes.Proof(
            steps=[p[1]])

    def p_theorem_proof_without_steps(
            self, p):
        """theorem_proof: PROOF"""
        p[0] = self._nodes.Proof(
            proof_keyword=True)

    def p_leaf_proof_with_facts_and_defs(
            self, p):
        """leaf_proof:
            by
                exprs
                def_keyword name_list
        """
        p[0] = self._nodes.By(
            only=p[1],
            facts=p[2],
            names=p[4])

    def p_leaf_proof_with_facts(
            self, p):
        """leaf_proof: by exprs"""
        p[0] = self._nodes.By(
            only=p[1],
            facts=p[2])

    def p_leaf_proof_with_defs(
            self, p):
        """leaf_proof:
            by
                def_keyword name_list
        """
        p[0] = self._nodes.By(
            only=p[1],
            names=p[3])

    def p_leaf_proof_obvious(
            self, p):
        """leaf_proof: OBVIOUS"""
        p[0] = self._nodes.Obvious()

    def p_leaf_proof_omitted(
            self, p):
        """leaf_proof: OMITTED"""
        p[0] = self._nodes.Omitted()

    def p_def_keyword(
            self, p):
        """def_keyword:
            | DEF
            | DEFS
        """
        p[0] = p[1]

    def p_by(
            self, p):
        """by: BY"""
        p[0] = False

    def p_by_only(
            self, p):
        """by: BY ONLY"""
        p[0] = True

    def p_proof_steps_append(
            self, p):
        """proof_steps:
            proof_steps proof_step
        """
        p[1].append(p[2])
        p[0] = p[1]

    def p_proof_steps_start(
            self, p):
        """proof_steps: proof_step"""
        p[0] = [p[1]]

    def p_proof_step_with_proof(
            self, p):
        """proof_step:
            | proof_goal PROOF leaf_proof
            | proof_goal leaf_proof
        """
        step_number, main = p[1]
        if p[2] == 'PROOF':
            proof = p[3]
            proof_keyword = True
        else:
            proof = p[2]
            proof_keyword = None
        p[0] = self._nodes.ProofStep(
            name=step_number,
            main=main,
            proof=proof,
            proof_keyword=proof_keyword)

    def p_proof_step_without_proof_pf_kw(
            self, p):
        """proof_step:
            proof_goal PROOF
        """
        step_number, main = p[1]
        p[0] = self._nodes.ProofStep(
            name=step_number,
            main=main,
            proof_keyword=True)

    def p_proof_step_without_proof(
            self, p):
        """proof_step:
            proof_goal
        """
        step_number, main = p[1]
        p[0] = self._nodes.ProofStep(
            name=step_number,
            main=main)

    def p_proof_step_main(
            self, p):
        """proof_goal:
            | assertion_proof_step
            | suffices_proof_step
            | case_proof_step
            | pick_proof_step
            | have_proof_step
            | take_proof_step
            | witness_proof_step
            | define_proof_step
            | instance_proof_step
            | use_proof_step
            | hide_proof_step
            | qed_proof_step
        """
        p[0] = p[1]

    def p_assertion_proof_step(
            self, p):
        """assertion_proof_step:
            | step_number sequent
            | step_number expr
        """
        p[0] = (p[1], p[2])

    def p_suffices_proof_step_sequent(
            self, p):
        """suffices_proof_step:
            step_number
                SUFFICES sequent
        """
        main = self._nodes.Suffices(
            predicate=p[3])
        p[0] = (p[1], main)

    def p_suffices_proof_step_expr(
            self, p):
        """suffices_proof_step:
            step_number
                SUFFICES expr
        """
        main = self._nodes.Suffices(
            predicate=p[3])
        p[0] = (p[1], main)

    def p_case_proof_step(
            self, p):
        """case_proof_step:
            step_number
                CASE
                    expr
                END_CASE
        """
        # because CASE is both undelimited
        # construct and proof step, and
        # `GeneratingParser` method
        # `_scopes_stack_append()` cannot
        # decide which
        main = self._nodes.ProofCase(
            expression=p[3])
        p[0] = (p[1], main)

    def p_pick_proof_step(
            self, p):
        """pick_proof_step:
            step_number
                PICK
                    exprs COLON
                        expr
                END_PICK
        """
        main = self._nodes.Pick(
            declarations=p[3],
            predicate=p[5])
        p[0] = (p[1], main)

    def p_have_proof_step(
            self, p):
        """have_proof_step:
            step_number
                HAVE expr
        """
        main = self._nodes.Have(
            expression=p[3])
        p[0] = (p[1], main)

    def p_proof_take(
            self, p):
        """take_proof_step:
            step_number
                TAKE exprs
        """
        main = self._nodes.Take(
            declarations=p[3])
        p[0] = (p[1], main)

    def p_proof_witness_step(
            self, p):
        """witness_proof_step:
            step_number
                WITNESS exprs
        """
        main = self._nodes.QuantifierWitness(
            expressions=p[3])
        p[0] = (p[1], main)

    def p_proof_define_step(
            self, p):
        """define_proof_step:
            step_number
                DEFINE operator_defs
                END_DEFINE
        """
        # The keyword `DEFINE` is
        # optional. The function
        # `_insert_proof_define_tokens()`
        # inserts `DEFINE` where it
        # has been omitted.
        main = self._nodes.Define(
            definitions=p[3])
        p[0] = (p[1], main)

    def p_instance_proof_step(
            self, p):
        """instance_proof_step:
            step_number instance
        """
        p[0] = (p[1], p[2])

    def p_proof_use_step(
            self, p):
        """use_proof_step:
            step_number use_unit
        """
        p[0] = (p[1], p[2])

    def p_proof_hide_step_full(
            self, p):
        """hide_proof_step:
            step_number
                HIDE
                    exprs
                    def_keyword name_list
        """
        main = self._nodes.Hide(
            facts=p[3],
            names=p[5])
        p[0] = (p[1], main)

    def p_proof_hide_step_with_facts(
            self, p):
        """hide_proof_step:
            step_number
                HIDE
                    exprs
        """
        main = self._nodes.Hide(
            facts=p[3])
        p[0] = (p[1], main)

    def p_proof_hide_step_with_defs(
            self, p):
        """hide_proof_step:
            step_number
                HIDE
                    def_keyword name_list
        """
        main = self._nodes.Hide(
            names=p[4])
        p[0] = (p[1], main)

    def p_qed_proof_step(
            self, p):
        """qed_proof_step:
            step_number
                QED
        """
        p[0] = (p[1], p[2])

    def p_proof_step_number(
            self, p):
        """step_number:
            | STEP_NUMBER
            | STEP_NUMBER_PLUS
            | STEP_NUMBER_ASTERISK
        """
        p[0] = p[1]

    def p_labeled_expr(
            self, p):
        """expr: op_app DOUBLE_COLON expr"""
        p[0] = self._nodes.OperatorApplication(
            operator=p[2],
            arguments=[p[1], p[3]])

    def p_general_identifier(
            self, p):
        """general_identifier: gid"""
        items = p[1]
        if len(items) == 1:
            op_app, = p[1]
            p[0] = op_app
            return
        if not items:
            raise ValueError(items)
        p[0] = self._nodes.SubexpressionReference(
            items=items)
            # NOTE: also `M!Op` where
            # `M == INSTANCE`
            # is represented here

    def p_gid_repeat_nullary_op_app(
            self, p):
        """gid:
            gid EXCLAMATION_NAME
        """
        name = p[2]
        p[1].append(name)
        p[0] = p[1]

    def p_gid_repeat_non_nullary_op_app(
            self, p):
        """gid:
            gid EXCLAMATION_NAME
                LPAREN exprs RPAREN
        """
        name = p[2]
        op_app = self._nodes.OperatorApplication(
            operator=name,
            arguments=p[4])
        p[1].append(op_app)
        p[0] = p[1]

    def p_gid_repeat_paren(
            self, p):
        """gid:
            gid EXCLAMATION
                LPAREN expr RPAREN
        """
        arguments = [p[4]]
        op_app = self._nodes.OperatorApplication(
            operator=None,
            arguments=arguments)
        p[1].append(op_app)
        p[0] = p[1]

    def p_gid_repeat_paren_exprs_2(
            self, p):
        """gid:
            gid EXCLAMATION
                LPAREN exprs_2 RPAREN
        """
        op_app = self._nodes.OperatorApplication(
            operator=None,
            arguments=p[4])
        p[1].append(op_app)
        p[0] = p[1]

    def p_gid_repeat_exclamation_symbol(
            self, p):
        """gid:
            gid exclamation_symbol
        """
        p[1].append(p[2])
        p[0] = p[1]

    def p_gid_repeat_exclamation_integer(
            self, p):
        """gid:
            gid EXCLAMATION DECIMAL_INTEGER
        """
        p[1].append(p[3])
        p[0] = p[1]

    def p_gid_start(
            self, p):
        """gid: op_app"""
        p[0] = [p[1]]

    def p_exclamation_symbol(
            self, p):
        """exclamation_symbol:
            | EXCLAMATION_AT
            | EXCLAMATION_COLON
            | EXCLAMATION_DOUBLE_LANGLE
            | EXCLAMATION_DOUBLE_RANGLE
        """
        p[0] = p[1]

    def p_symbol_at(
            self, p):
        """expr: AT"""
        p[0] = p[1]


class OperatorPrecedenceGrammar(Grammar):
    """LR(1) with operator-precedence table."""

    def __init__(
            self
            ) -> None:
        self.operator_precedence = [
            ('nonassoc', 'DOUBLE_COLON'),
                # labels
            ('nonassoc', 'OP_1_INFIX'),
            ('nonassoc', 'OP_2_INFIX'),
            ('left', 'OP_3_LEFT'),
            ('nonassoc', 'OP_4_PREFIX'),
            ('nonassoc', 'OP_4_BEFORE'),
            ('nonassoc', 'EQ'),
                # op_5_eq_infix
            ('left', 'BSLASH_IN'),
                # 'op_5_bslash_in_between',
            ('left', 'OP_5_LEFT'),
            ('nonassoc', 'OP_5_INFIX'),
            ('left', 'OP_6_LEFT'),
            ('nonassoc', 'OP_7_INFIX'),
            ('nonassoc', 'OP_8_BEFORE'),
            ('nonassoc', 'OP_8_INFIX'),
            ('left', 'OP_8_LEFT'),
            ('nonassoc', 'OP_9_BEFORE'),
            ('nonassoc', 'OP_9_INFIX'),
            ('left', 'OP_9_LEFT'),
            ('left', 'OP_10_LEFT'),
            ('nonassoc', 'OP_10_INFIX'),
            ('nonassoc', 'set_of_tuples'),
            ('left', 'CARTESIAN'),
                # 'set_of_tuples_between',
            ('left', 'OP_11_LEFT'),
            ('left', 'DASH'),
                # 'op_11_dash_left',
                # 'op_12_dash_before',
            ('nonassoc', 'OP_12_PREFIX'),
                # for `DASH_DOT`
            ('nonassoc', 'OP_13_INFIX'),
            ('left', 'OP_13_LEFT'),
            ('nonassoc', 'OP_14_INFIX'),
            ('nonassoc', 'OP_15_POSTFIX'),
            ('nonassoc', 'OP_15_AFTER'),
            ('nonassoc', 'LBRACKET'),
            ('nonassoc', 'DOT')]
        super().__init__()

    def p_expr_binary_operator(
            self, p):
        """expr:
            | expr OP_1_INFIX expr
            | expr OP_2_INFIX expr
            | expr OP_3_LEFT expr
            | expr EQ expr
            | expr BSLASH_IN expr
            | expr OP_5_LEFT expr
            | expr OP_5_INFIX expr
            | expr OP_6_LEFT expr
            | expr OP_7_INFIX expr
            | expr OP_8_INFIX expr
            | expr OP_8_LEFT expr
            | expr OP_9_INFIX expr
            | expr OP_9_LEFT expr
            | expr OP_10_LEFT expr
            | expr OP_10_INFIX expr
            | expr OP_11_LEFT expr
            | expr DASH expr
            | expr OP_13_INFIX expr
            | expr OP_13_LEFT expr
            | expr OP_14_INFIX expr
        """
        op_lexeme = p[2]
        args = [p[1], p[3]]
        p[0] = self._nodes.OperatorApplication(
            operator=op_lexeme,
            arguments=args)

    def p_expr_prefix_operator(
            self, p):
        """expr:
            | OP_4_PREFIX expr
            | OP_4_BEFORE expr
            | OP_8_BEFORE expr
            | OP_9_BEFORE expr
            | DASH expr
            | OP_12_PREFIX expr
        """
        p[0] = self._nodes.OperatorApplication(
            operator=p[1],
            arguments=[p[2]])

    def p_expr_postfix_operator(
            self, p):
        """expr:
            | expr OP_15_POSTFIX
            | expr OP_15_AFTER
        """
        p[0] = self._nodes.OperatorApplication(
            operator=p[2],
            arguments=[p[1]])

    def p_expr_set_of_tuples(
            self, p):
        """expr: set_of_tuples"""
        op_lexeme, *args = p[1]
        p[0] = self._nodes.OperatorApplication(
            operator=op_lexeme,
            arguments=args)

    def p_set_of_tuples_repeat(
            self, p):
        """set_of_tuples:
            set_of_tuples CARTESIAN expr
        """
        p[1].append(p[3])
        p[0] = p[1]

    def p_set_of_tuples_start(
            self, p):
        """set_of_tuples:
            expr CARTESIAN expr
        """
        p[0] = [p[2], p[1], p[3]]

    def p_expr_function_application(
            self, p):
        """expr: function_application"""
        p[0] = p[1]



"""Tests for vertical expressions with `|`."""
import tla._ast as _ast
import tla._lre as _lr


source = r'''
---- MODULE example ----

A == | B
     | C

========================
'''


def test_vertical_pipes():
    parser = _lr.Parser()
    tree = parser.parse(source)
    _ast.pprint_tree(tree)


if __name__ == '__main__':
    test_vertical_pipes()



"""Operator fixity and associativity.

These definitions can then be shared between LR parsers.
"""
import itertools as _itr
import typing as _ty

import tla._utils as _utils


PrecRange: _ty.TypeAlias = tuple[int, int]


KEYWORDS = {
    'ACTION',
    'ASSUME',
    'ASSUMPTION',
    'AXIOM',
    'BOOLEAN',
    'BY',
    'CASE',
    'CHOOSE',
    'CONSTANT',
    'CONSTANTS',
    'COROLLARY',
    'DEF',
    'DEFINE',
    'DEFS',
    'DOMAIN',
    'ELSE',
    'ELIF',
        # added to TLA
    'ENABLED',
    'EXCEPT',
    'EXTENDS',
    'FALSE',
    'HAVE',
    'HIDE',
    'IF',
    'IN',
    'INSTANCE',
    'LAMBDA',
    'LEMMA',
    'LET',
    'LOCAL',
    'MODULE',
    'NEW',
    'OBVIOUS',
    'OMITTED',
    'ONLY',
    'OTHER',
    'PICK',
    'PROOF',
    'PROPOSITION',
    'PROVE',
    'QED',
    'RECURSIVE',
    # 'SF_',
        # regex used
    'STATE',
    'STRING',
    'SUBSET',
    'SUFFICES',
    'TAKE',
    'TEMPORAL',
    'THEN',
    'THEOREM',
    'TRUE',
    'UNCHANGED',
    'UNION',
    'USE',
    'VARIABLE',
    'VARIABLES',
    # 'WF_',
        # regex used
    'WITH',
    'WITNESS',
    }
_utils.assert_len(KEYWORDS, 58)


def _collect_fixal(
        *fixities:
            str
        ) -> set[str]:
    """Return operator names."""
    sets = _itr.chain.from_iterable(
        TLA_OPERATOR_FIXITY.get(fixity, set())
        for fixity in fixities)
    return set().union(sets)


def _bslash(
        suffix:
            str
        ) -> str:
    r"""Prefix with `\`."""
    return f'\\{suffix}'


_FIXITIES: _ty.Final = {
    'prefix', 'before',
    'infix', 'left', 'between', 'right',
    'postfix', 'after',
    }
_utils.assert_len(_FIXITIES, 8)
# Table 6 on page 271 of
# the book "Specifying Systems"
LEXEME_PRECEDENCE: _ty.Final = {
    # symbol: (start, end, fixity, associativity)
    # where:
    # - `start..end`: is the set of precedence
    #   levels of the operator
    # - `fixity`:
    #   - prefix (not directly nestable)
    #   - before (nestable prefix)
    #   - infix (not directly nestable)
    #   - left (nestable, left-associative)
    #   - right (nestable, right-associative)
    #   - between (listy, parsed as separator)
    #   - postfix (not directly nestable)
    #   - after (nestable postfix)
    #
    #   (nonfix operators are user-defined
    #    and represented by alphanumeric
    #    identifiers, so not in this `dict`)
    #
    # logic
    '=>':
        (1, 1),
    '<=>':
        (2, 2),
    '/\\':
        (3, 3),
    r'\/':
        (3, 3),
    '~':
        (4, 4),
    '=':
        (5, 5),
    '#':
        (5, 5),
    '/=':
        (5, 5),
    # set theory
    'SUBSET':
        (8, 8),
    'UNION':
        (8, 8),
    'DOMAIN':
        (9, 9),
    r'\subseteq':
        (5, 5),
    r'\in':
        (5, 5),
    r'\notin':
        (5, 5),
    '\\':
        (8, 8),
    r'\cap':
        (8, 8),
    r'\cup':
        (8, 8),
    r'\X':
        (10, 13),
    # action operators
    "'":
        (15, 15),
    'UNCHANGED':
        (4, 15),
    r'\cdot':
        (5, 14),
    'ENABLED':
        (4, 15),
    # modal operators
    '[]':
        (4, 15),
    '<>':
        (4, 15),
    '~>':
        (2, 2),
    '-+->':
        (2, 2),
    # user-defined operators
    '^':
        (14, 14),
    '/':
        (13, 13),
    '*':
        (13, 13),
    '-.':
        (12, 12),
        # within operator signatures
    # '-':
    #    (11, 11),
        # `op_11_dash_left` in
        # `positioning`, within
        #  the class `Grammar`
    # '-':
    #    (12, 12),
        # `op_12_dash_before` in
        # `positioning`, within
        # the class `Grammar`
    '+':
        (10, 10),
    '^+':
        (15, 15),
    '^*':
        (15, 15),
    '^#':
        (15, 15),
    '<':
        (5, 5),
    '=<':
        (5, 5),
    '<=':
        (5, 5),
    '>':
        (5, 5),
    '>=':
        (5, 5),
    '..':
        (9, 9),
    '...':
        (9, 9),
    '|':
        (10, 11),
    '||':
        (10, 11),
    '&':
        (13, 13),
    '&&':
        (13, 13),
    '$':
        (9, 13),
    '$$':
        (9, 13),
    '??':
        (9, 13),
    '%':
        (10, 11),
    '%%':
        (10, 11),
    '##':
        (9, 13),
    '++':
        (10, 10),
    '--':
        (11, 11),
    '**':
        (13, 13),
    '//':
        (13, 13),
    '^^':
        (14, 14),
    '@@':
        (6, 6),
    '!!':
        (9, 13),
    '|-':
        (5, 5),
    '|=':
        (5, 5),
    '-|':
        (5, 5),
    '=|':
        (5, 5),
    '<:':
        (7, 7),
    ':>':
        (7, 7),
    ':=':
        (5, 5),
    '::=':
        (5, 5),
    '(+)':
        (10, 10),
    '(-)':
        (11, 11),
    '(.)':
        (13, 13),
    '(/)':
        (13, 13),
    r'(\X)':
        (13, 13),
    } | _utils.map_keys(_bslash, {
        'uplus':
            (9, 13),
        'sqcap':
            (9, 13),
        'sqcup':
            (9, 13),
        'div':
            (13, 13),
        'wr':
            (9, 14),
        'star':
            (13, 13),
        'circ':
            (13, 13),
        'o':
            (13, 13),
        'bigcirc':
            (13, 13),
        'bullet':
            (13, 13),
        'prec':
            (5, 5),
        'succ':
            (5, 5),
        'preceq':
            (5, 5),
        'succeq':
            (5, 5),
        'sim':
            (5, 5),
        'simeq':
            (5, 5),
        'll':
            (5, 5),
        'gg':
            (5, 5),
        'asymp':
            (5, 5),
        'subset':
            (5, 5),
        'supset':
            (5, 5),
        'supseteq':
            (5, 5),
        'approx':
            (5, 5),
        'cong':
            (5, 5),
        'sqsubset':
            (5, 5),
        'sqsubseteq':
            (5, 5),
        'sqsupset':
            (5, 5),
        'sqsupseteq':
            (5, 5),
        'doteq':
            (5, 5),
        'propto':
            (5, 5),
        })
_utils.assert_len(
    LEXEME_PRECEDENCE, 101)
TLA_OPERATOR_FIXITY: _ty.Final = dict(
    prefix={
        'UNCHANGED',
        '-.',  # in operator signatures
        },
    before={
        '~', '[]', '<>',
        # the dash `-` appears
        # within the class `Grammar`
        'DOMAIN',
        'ENABLED',
        'SUBSET',
        'UNION',
        },
    infix={
        '=>', '<=>', '~>', '-+->',
        '=', '#', '/=',
        '^', '/',
        '<', '<=', '=<', '>=', '>',
        '..', '...', '%',
        '\\', '//', '^^', '!!',
        '|-', '-|',
        '|=', '=|',
        '<:', ':>',
        ':=', '::=',
        '(/)',
        }.union(map(_bslash, [
            'approx',
            'asymp',
            'cong',
            'div',
            'doteq',
            'gg',
            'll',
            'notin',
            'prec',
            'preceq',
            'propto',
            'sim',
            'simeq',
            'sqsubset',
            'sqsubseteq',
            'sqsupset',
            'sqsupseteq',
            'subset',
            'subseteq',
            'succ',
            'succeq',
            'supset',
            'supseteq',
            'wr',
            ])),
    left={
        '/\\', r'\/',
        r'\cap', r'\cup',
        '*', '+',
        # the dash `-` appears
        # within the class `Grammar`
        '|', '||',
        '&', '&&',
        '$', '$$',
        '??',
        '%%', '##',
        '++', '--', '**',
        '@@',
        r'\cdot',
        '(+)', '(-)', '(.)', r'(\X)',
        }.union(map(_bslash, [
            'bigcirc',
            'bullet',
            'circ',
            'o',
            'sqcap',
            'sqcup',
            'star',
            'uplus',
            ])),
        # the dot `.` appears
        # directly in the grammar
    between=set(map(_bslash, [
        'in',
            # The operator `\in`
            # cannot be nested inside
            # expressions. It is here
            # to represent declarations.
        'X',
            # The operator `\X`
            # is described in
            # Section 15.2.1 on
            # page 284 of the book
            # "Specifying Systems".
        ])),
    postfix={
        "'",
        },
    after={
        '^+', '^*', '^#',
        },
    )
_utils.assert_len(
    TLA_OPERATOR_FIXITY, 7)
# Lexemes of vertical operators.
VERTICAL_LEXEMES: _ty.Final = {
    '/\\', r'\/',
    # '&', '&&',
    # '*', '**',
    # '%', '%%',
    '|',  # '||',
    # '+', '++',
    # '@@',
    # '^^',
    # '--',
    # '!!',
    # '##',
    # '??'
    }
_utils.assert_len(
    VERTICAL_LEXEMES, 3)
INFIXAL: _ty.Final = _collect_fixal(
    'infix', 'left', 'right', 'between')
POSTFIXAL: _ty.Final = _collect_fixal(
    'postfix', 'after')
PREFIXAL: _ty.Final = _collect_fixal(
    'prefix', 'before')


def check_lexprec(
        ) -> None:
    """Assert properties of operator tables."""
    for prec in LEXEME_PRECEDENCE.values():
        _check_precedence(prec)
    # assert pairwise disjoint sets
    accum = set()
    items = TLA_OPERATOR_FIXITY.items()
    for fixity, operators in items:
        if accum & operators:
            raise ValueError(
                fixity, operators)
        accum.update(operators)

def _check_precedence(
        prec:
            PrecRange
        ) -> None:
    """Assert precedence is correct."""
    if len(prec) != 2:
        raise ValueError(prec)
    start, end = prec
    if 1 > start:
        raise ValueError(start)
    if start > end:
        raise ValueError(start, end)



"""Generation and coalescing of tokens.

This is a layer between lexer and parser.
"""
import collections.abc as _abc
import functools as _ft

import tla._langdef as _opdef
import tla._lex as _lex
import tla._utils as _utils


_PREFIXAL = _opdef.PREFIXAL
_INFIXAL = _opdef.INFIXAL
_POSTFIXAL = _opdef.POSTFIXAL
# Tokens that are step numbers.
STEP_NUMBER_TOKENS = {
    'STEP_NUMBER',
    'STEP_NUMBER_PLUS',
    'STEP_NUMBER_ASTERISK',
    }
_utils.assert_len(
    STEP_NUMBER_TOKENS, 3)
# A step number that follows any
# of these tokens is the start
# of a step. Otherwise, step numbers
# are identifiers (converted to
# token `NAME`).
_TOKENS_BEFORE_STEP_START = {
    # operators: POSTFIX
    'BOOLEAN',
    'FALSE',
    'OBVIOUS',
    'OMITTED',
    'PROOF',
    'QED',
    'STRING',
    'TRUE',
    'BINARY_INTEGER',
    'DECIMAL_INTEGER',
    'FLOAT',
    'HEXADECIMAL_INTEGER',
    'MULTILINE_STRING_LITERAL',
    'NAME',
    'OCTAL_INTEGER',
    'STEP_NUMBER',
    'STEP_NUMBER_ASTERISK',
    'STEP_NUMBER_PLUS',
    'STRING_LITERAL',
    'RPAREN',
    'RBRACE',
    'RBRACKET',
    'DOUBLE_RANGLE',
    # coalesced tokens created by
    # `join_as_one_token()`
    # which are part of
    # subexpression references
    'EXCLAMATION_AT',
    'EXCLAMATION_COLON',
    'EXCLAMATION_DOUBLE_LANGLE',
    'EXCLAMATION_DOUBLE_RANGLE',
    'EXCLAMATION_NAME',
    }
_utils.assert_len(
    _TOKENS_BEFORE_STEP_START, 28)
MODULE_SCOPE_TOKENS = STEP_NUMBER_TOKENS | {
    'DEF_SIG',
    'EQ_LINE',
    'DASH_LINE',
    'VARIABLE',
    'VARIABLES',
    'CONSTANT',
    'CONSTANTS',
    'RECURSIVE',
    'ASSUMPTION',
    'ASSUME',
    'AXIOM',
    'THEOREM',
    'PROPOSITION',
    'PROOF',
    'PROVE',
    'LEMMA',
    'COROLLARY',
    'BY',
    'DEF',
    'DEFS',
    'OBVIOUS',
    'OMITTED',
    'USE',
    'HIDE',
    'INSTANCE',
        # `INSTANCE` occurs either:
        # - in module scope
        # - after `==`
        # - after `STEP_NUMBER`
    }
_utils.assert_len(
    MODULE_SCOPE_TOKENS,
    25 + len(STEP_NUMBER_TOKENS))
# Tokens that close a `DEFINE` section.
_DEFINE_CLOSING_TOKENS = (
    MODULE_SCOPE_TOKENS - {
        'DEF_SIG', 'INSTANCE'})
_utils.assert_len(
    _DEFINE_CLOSING_TOKENS,
    -2 + len(MODULE_SCOPE_TOKENS))
_BEFORE_POSTFIXAL = {
    'DEF',
    'DEFS',
    'COMMA',
    'LPAREN',
    'WITH',
    'LARROW',
    'EXCLAMATION',
    }
_utils.assert_len(
    _BEFORE_POSTFIXAL, 7)
_AFTER_PREFIXAL = MODULE_SCOPE_TOKENS | {
    'COMMA',
    'RPAREN',
    'LARROW',
    'EXCLAMATION',
    }
_utils.assert_len(
    _AFTER_PREFIXAL,
    4 + len(MODULE_SCOPE_TOKENS))


def preparse(
        tokens:
            _abc.Iterable
        ) -> _abc.Iterable:
    """Prepare tokens for parser."""
    tokens = insert_def_starts_tokens(tokens)
    tokens = swap_local_keyword(tokens)
    tokens = join_as_one_token(tokens)
    tokens = switch_keywords_after_dot(tokens)
        # `L.WITH &` needs switching `WITH` to
        # `NAME` token before checking the
        # fixal operator `&`
    tokens = switch_fixal_operators(tokens)
    tokens = switch_step_identifiers(tokens)
        # `BY DEF + <1>2` needs switching fixal
        # operators to `NAME`, before switching
        # step identifiers
    tokens = insert_proof_define_tokens(tokens)
    tokens = end_define_proof_steps(tokens)
    return tokens


def insert_def_starts_tokens(
        tokens:
            _abc.Iterable
        ) -> list:
    """Delimit start of definitions."""
    return _utils.pipe(
        tokens,
        _utils.reverse,
        _insert_def_sig_tokens_nonfix,
        _insert_def_sig_tokens_fixal,
        _utils.reverse)


def _insert_def_sig_tokens_fixal(
        tokens:
            _abc.Iterable
        ) -> _abc.Iterable:
    r"""Insert `DEF_SIG` tokens.

    - `name infixal name ==`
    - `name ==`
    - `-. name ==`
    - `name postfixal ==`
    """
    stack = list()
    def make_def_sig(
            token):
        return _utils.make_lex_token(
            'DEF_SIG',
            row=token.row,
            column=token.column,
            start=token.start)
    last_tokens = list()
    for token in tokens:
        # prefix-operator definition
        is_prefix_def = (
            len(last_tokens) >= 2 and
            last_tokens[-2].symbol == 'DOUBLE_EQ' and
            last_tokens[-1].symbol == 'NAME' and
            token.value in _PREFIXAL)
        if is_prefix_def:
            yield token
            yield make_def_sig(token)
        # infix-operator definition
        is_infix_def = (
            len(last_tokens) >= 3 and
            last_tokens[-3].symbol == 'DOUBLE_EQ' and
            last_tokens[-2].symbol == 'NAME' and
            last_tokens[-1].value in _INFIXAL and
            token.symbol == 'NAME')
        if is_infix_def:
            yield token
            yield make_def_sig(token)
        # postfix-operator definition
        is_postfix_def = (
            len(last_tokens) >= 2 and
            last_tokens[-2].symbol == 'DOUBLE_EQ' and
            last_tokens[-1].value in _POSTFIXAL and
            token.symbol == 'NAME')
        if is_postfix_def:
            yield token
            yield make_def_sig(token)
        # nullary-operator definition
        is_nullary_def = (
            len(last_tokens) >= 2 and
            last_tokens[-2].symbol == 'DOUBLE_EQ' and
            last_tokens[-1].symbol == 'NAME' and
            not is_prefix_def and
            token.value not in _INFIXAL)
        if is_nullary_def:
            yield make_def_sig(token)
            yield token
        is_def = (
            is_prefix_def or
            is_infix_def or
            is_postfix_def or
            is_nullary_def)
        if not is_def:
            yield token
        last_tokens.append(token)
    if not stack:
        return
    raise ValueError(stack)


def _insert_def_sig_tokens_nonfix(
        tokens:
            _abc.Sequence
        ) -> _abc.Iterable:
    r"""Insert `DEF_SIG` tokens.

    Cases:
    - `name(...) ==`
    - `name[...] ==` (nesting of definition
        signatures can occur here, e.g.,
        `f[x \in LET -. r == r IN {-1}] == x`)

    `LOCAL` can precede any of these cases.

    @param tokens:
        sequence of tokens reversed from
        how they appear in the TLA+ source,
        i.e., in the lexer's output
    @return:
        sequence of tokens,
        in the same direction as the input
        (i.e., reversed from their order
         in the TLA+ source)
    """
    stack = list()
    def make_def_sig(
            token):
        return _utils.make_lex_token(
            'DEF_SIG',
            row=token.row,
            column=token.column,
            start=token.start)
    last_token = _utils.make_lex_token(None)
    for token in tokens:
        ts = token.symbol
        if (ts in ('RPAREN', 'RBRACKET') and
                last_token.symbol == 'DOUBLE_EQ'):
            stack.append(2)
            yield token
        elif not stack:
            yield token
        elif (ts in ('LPAREN', 'LBRACKET') and
                stack[-1] > 0):
            stack[-1] -= 1
            yield token
        elif (ts in ('RPAREN', 'RBRACKET') and
                stack[-1] > 0):
            stack[-1] += 1
            yield token
        elif ts == 'NAME' and stack[-1] == 1:
            stack.pop()
            yield token
            yield make_def_sig(token)
        elif stack[-1] != 0:
            yield token
        else:
            yield token
        last_token = token
    if not stack:
        return
    raise ValueError(stack)


def swap_local_keyword(
        tokens:
            _abc.Iterable
        ) -> _abc.Iterable:
    """Swap `LOCAL DEF_SIG` to `DEF_SIG LOCAL`."""
    previous = None
    for token in tokens:
        ts = token.symbol
        if ts == 'LOCAL':
            previous = token
            continue
        swap = (
            previous is not None and
            token.symbol == 'DEF_SIG')
        if swap:
            yield token
            yield previous
        elif previous is not None:
            yield previous
            yield token
        else:
            yield token
        if previous is not None:
            previous = None
    if previous is not None:
        yield previous


def insert_proof_define_tokens(
        tokens:
            _abc.Iterable
        ) -> _abc.Iterable:
    """Insert any missing `DEFINE` tokens."""
    last_token = _utils.make_lex_token(None)
    for token in tokens:
        if (token.symbol == 'DEF_SIG' and
                last_token.symbol in
                    STEP_NUMBER_TOKENS):
            # column information is not needed,
            # because `DEFINE` cannot follow
            # a vertical item
            yield _utils.make_lex_token('DEFINE')
        last_token = token
        yield token


def end_define_proof_steps(
        tokens:
            _abc.Iterable
        ) -> _abc.Iterable:
    """Add `END_DEFINE` at end of `DEFINE`."""
    inside_define_step = False
    previous = _utils.make_lex_token(None)
    for token in tokens:
        ts = token.symbol
        pt = previous.symbol
        module_scope_instance = (
            ts == 'INSTANCE' and
            pt != 'DOUBLE_EQ')
        is_closing = (
            ts in _DEFINE_CLOSING_TOKENS or
            module_scope_instance)
        add_token = (
            inside_define_step and
            is_closing)
        if add_token:
            inside_define_step = False
            # column information is not needed,
            # because `END_DEFINE` ends also
            # vertical items, as a
            # non-blankspace token
            yield _utils.make_lex_token('END_DEFINE')
        if ts == 'DEFINE':
            inside_define_step = True
        previous = token
        yield token


def switch_step_identifiers(
        tokens:
            _abc.Iterable
        ) -> _abc.Iterable:
    """Convert step identifiers to `NAME`.

    For occurrences of step identifiers
    within leaf proofs (`BY` proofs).
    """
    last_token = _utils.make_lex_token(None)
    for token in tokens:
        switch = (
            token.symbol in
                STEP_NUMBER_TOKENS and
            last_token.symbol not in
                _TOKENS_BEFORE_STEP_START and
            last_token.value not in
                _POSTFIXAL)
        if switch:
            token = _utils.copy_lex_token(
                token,
                symbol='NAME')
        last_token = token
        yield token


def join_as_one_token(
        tokens:
            _abc.Iterable
        ) -> _abc.Iterable:
    """Map `<<TOK1, TOK2>>` to `<<TOK1_TOK2>>`.

    Coalescing:

    - `<<EXCLAMATION, AT>>`
      to `<<EXCLAMATION_AT>>`

    - `<<EXCLAMATION, COLON>>`
      to `<<EXCLAMATION_COLON>>`

    - `<<EXCLAMATION, DOUBLE_LANGLE>>`
      to `<<EXCLAMATION_DOUBLE_LANGLE>>`

    - `<<EXCLAMATION, NAME>>`
      to `<<EXCLAMATION_NAME>>`

    - `<<EXCLAMATION, DOUBLE_RANGLE>>`
      to `<<EXCLAMATION_DOUBLE_RANGLE>>`
    """
    toks = {
        'AT',
        'COLON',
        'DOUBLE_LANGLE',
        'DOUBLE_RANGLE',
        'NAME',
        }
    pairs = {
        f'EXCLAMATION_{s}'
        for s in toks}
    first = {'EXCLAMATION'}
    notoken = _utils.make_lex_token(None)
    last_token = notoken
    for token in tokens:
        a = last_token.symbol
        b = token.symbol
        token_type = f'{a}_{b}'
        if token_type in pairs:
            if token_type.startswith(
                    'EXCLAMATION'):
                token_value = token.value
            else:
                raise ValueError(token_type)
            yield _utils.copy_lex_token(
                last_token,
                symbol=token_type,
                value=token_value)
            last_token = notoken
            continue
        elif last_token.symbol in first:
            yield last_token
            last_token = token
        if token.symbol in first:
            last_token = token
            continue
        last_token = notoken
        yield token
    # This case can arise when the
    # start symbol of the grammar is
    # an expression. An expression
    # can end in a `NAME` token.
    if last_token != notoken:
        yield last_token


def switch_fixal_operators(
        tokens:
            _abc.Iterable
        ) -> list:
    """Convert fixal to `NAME`.

    For operators as arguments, for example:
    `Op(+, 1, 2)`

    The `+` is converted from an operator
    to `NAME`.
    """
    forward_scan = _ft.partial(
        _switch_scan,
        current=_INFIXAL | _POSTFIXAL,
        previous=_BEFORE_POSTFIXAL)
    reverse_scan = _ft.partial(
        _switch_scan,
        current=_PREFIXAL,
        previous=_AFTER_PREFIXAL)
    return _utils.pipe(
        tokens,
        forward_scan,
        _utils.reverse,
        reverse_scan,
        _switch_fixal_exclamation,
        _utils.reverse)


def switch_keywords_after_dot(
        tokens:
            _abc.Iterable
        ) -> _abc.Iterable:
    """Map `DOT KEYWORD` to `DOT NAME`."""
    keywords = set(_opdef.KEYWORDS)
    keywords.update(['SF_', 'WF_'])
    return _switch_scan(
        tokens,
        current=keywords,
        previous={'DOT'})


def _switch_scan(
        tokens:
            _abc.Iterable,
        current:
            set[str],
        previous:
            set[str]
        ) -> _abc.Iterable:
    previous_token = _utils.make_lex_token(None)
    for token in tokens:
        switch = (
            token.value in current and
            previous_token.symbol in previous)
        if switch:
            token = _utils.copy_lex_token(
                token,
                symbol='NAME')
            assert token.symbol == 'NAME', token
        previous_token = token
        yield token


def _switch_fixal_exclamation(
        tokens:
            _abc.Iterable
        ) -> _abc.Iterable:
    """Map fixal before `(...)!`, to `NAME`."""
    stack = list()
    last_token = None
    for token in tokens:
        ts = token.symbol
        if (ts == 'RPAREN' and
                last_token is not None and
                last_token.symbol.startswith(
                    'EXCLAMATION')):
            stack.append(1)
            yield token
        elif not stack:
            yield token
        elif ts == 'RPAREN':
            stack[-1] += 1
            yield token
        elif ts == 'LPAREN':
            stack[-1] -= 1
            yield token
        elif stack[-1] > 0:
            yield token
        elif (ts.startswith('OP_') and
                stack[-1] == 0):
            stack.pop()
            yield _utils.copy_lex_token(
                token,
                symbol='NAME')
        elif stack[-1] == 0:
            # A(...)!
            # already `NAME`, not fixal
            stack.pop()
            yield token
        else:
            yield token
        last_token = token
    if not stack:
        return
    raise ValueError(stack)


def filter_comments(
        tokens:
            _abc.Iterable
        ) -> _abc.Iterator:
    """Omit tokens that are comments."""
    comments = _lex.COMMENT_TOKENS
    for token in tokens:
        if token.symbol in comments:
            continue
        yield token


def filter_outside_module(
        tokens:
            _abc.Iterable
        ) -> _abc.Iterable:
    """Omit tokens before and after root module."""
    tokens = list(tokens)
    module_count = 0
    for i, token in enumerate(tokens):
        if i + 1 >= len(tokens):
            next_symbol = None
        else:
            next_symbol = tokens[i + 1].symbol
        module_starts = (
            token.symbol == 'DASH_LINE' and
            next_symbol == 'MODULE')
        if module_starts:
            module_count += 1
        module_ends = (
            token.symbol == 'EQ_LINE')
        if module_ends and module_count > 1:
            module_count -= 1
        elif module_ends and module_count == 1:
            yield token
            break
        if module_count > 0:
            yield token



"""Rearranging blankspace within TLA+ source.

How to take TLA+ source, parse it, and then
print the syntax tree as TLA+ source, rearranged
in a specific way according to formatting rules.
"""
import tla


TLA_TEXT = r'''
----
MODULE example
----
(* An example module. *)
EXTENDS
Integers A == LET
x ==1 y== 2 IN x+ y THEOREM
A*2 = 2* (1+2 ) BY DEF A, *, +
====
'''


def format_tla():
    """Example of formatting, with comments preserved."""
    formatted = tla.pformat_tla(TLA_TEXT)
    print(formatted)


if __name__ == '__main__':
    format_tla()



"""How to parse a TLA+ expression."""
import tla


expr = r'''
    \/ /\ x = 1
       /\ x' = 2

    \/ /\ x = 2
       /\ x' = 1
    '''


def parse_expr():
    """Parse a TLA+ expression."""
    tree = tla.parse_expr(expr)
    print(tree)


def parse_expr_and_format():
    """Parse, format, print a TLA+ expression."""
    tree = tla.parse_expr(expr)
    text = tla.pformat_ast(tree)
    print(text)


if __name__ == '__main__':
    parse_expr()
    parse_expr_and_format()



"""Test Boolean conditional."""
import tla._lre as _lr


def test_elif_conditional():
    module = '''
    ---- MODULE elif_test ----
    A == IF TRUE THEN
            1
        ELIF TRUE THEN
            2
        ELIF FALSE THEN
            3
        ELSE
            4
    ==========================
    '''
    parser = _lr.Parser()
    tree = parser.parse(module)
    print(tree)


if __name__ == '__main__':
    test_elif_conditional()
